package gov.va.med.vistalink.management;

import gov.va.med.vistalink.adapter.spi.ConnectorInfoVO;
import gov.va.med.vistalink.adapter.spi.VistaLinkRuntimeInfo;
import gov.va.med.vistalink.institution.InstitutionMapManager;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Map;

import javax.management.InstanceNotFoundException;
import javax.management.ObjectName;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.Logger;

import weblogic.management.MBeanHome;
import weblogic.management.RemoteMBeanServer;

/**
 * WebLogic-specific console helper functions.
 * 
 * @author Application Modernization - Foundations Team
 * @version 1.5.0.026
 */
public class ConsoleHelper {

	/**
	 * explicitly set the encoding format for console pages
	 * @va.exclude
	 */
	public static final String ENCODE_FORMAT = "UTF-8";

	/**
	 * explicitly set the response content type for console pages
	 * @va.exclude
	 */
	public static final String RESPONSE_CONTENT_TYPE = "text/html; charset=" + ENCODE_FORMAT;

	private static final Logger logger = Logger.getLogger(ConsoleHelper.class);

	/**
	 * Helper function to return current institution mapings
	 * @param serverName Server to return mappings for
	 * @return Map of institution mappings, returned from InstitutionMapManager MBean on specified server.
	 * @throws MBeanHomeNotFoundException
	 * @throws Exception
	 * @va.exclude
	 */
	public static Map getInstitutionInfo(String serverName) throws MBeanHomeNotFoundException, Exception {

		Map returnVal = null;
		String domainName = null;

		try {

			MBeanHome remoteHome = ConsoleHelper.getRemoteHome(serverName);
			RemoteMBeanServer remoteMBeanServer = remoteHome.getMBeanServer();
			domainName = remoteHome.getDomainName();
			String canonicalInstMapMgrName = InstitutionMapManager.getServerSpecificCanonicalObjectName(domainName,
					serverName);
			logger.debug("about to make remote JMX InstitutionMapManager getMappings call.");
			returnVal = (Map) remoteMBeanServer.invoke(new ObjectName(canonicalInstMapMgrName), "getMappings",
					new Object[0], new String[0]);
			if (logger.isDebugEnabled()) {
				logger.debug("Returned with institution list with " + returnVal.size() + " entries.");
			}
		} catch (MBeanHomeNotFoundException e) {
			// don't log
			throw e;
		} catch (InstanceNotFoundException e) {
			logger.warn(e);
			// could not find the 1-per-server InstitutionMapManager MBean
			throw new VistaLinkMBeanNotFoundException(e);

		} catch (Exception e) {
			logger.error("Could not get Institution Info: " + e.getMessage());
			throw e;
		}

		return returnVal;
	}

	/**
	 * Retrieves a list of currently deployed (partially or fully) VistaLink connectors, with summary info for each.
	 * @param serverName name of server to retrieve connector list for for.
	 * @return Map with key=ConnectorComponentRuntimeMBean canonical object name for the connector,
	 *         value=ConnectorInfoVO object with summary information for the particular connector.
	 * @throws Exception
	 * @va.exclude
	 */
	public static Map getConnectorList(String serverName) throws Exception {

		String domainName = null;
		Map returnVal = null;

		try {
			MBeanHome remoteHome = ConsoleHelper.getRemoteHome(serverName);
			RemoteMBeanServer remoteMBeanServer = remoteHome.getMBeanServer();
			domainName = remoteHome.getDomainName();
			String canonicalVljMcfMonitorName = VistaLinkRuntimeInfo.getServerSpecificCanonicalObjectName(domainName,
					serverName);
			logger.debug("about to make remote JMX retrieveConnectorList call.");
			returnVal = (Map) remoteMBeanServer.invoke(new ObjectName(canonicalVljMcfMonitorName),
					"retrieveConnectorList", new Object[0], new String[0]);
			if (logger.isDebugEnabled()) {
				logger.debug("Returned with connector list with " + returnVal.size() + " entries.");
			}
		} catch (MBeanHomeNotFoundException e) {
			// don't log
			throw e;
		} catch (InstanceNotFoundException e) {
			logger.warn(e);
			// could not find the 1-per-server VistaLinkRuntimeInfoMBean
			throw new VistaLinkMBeanNotFoundException(e);
		} catch (Exception e) {
			logger.error(e);
			throw e;
		}

		return returnVal;
	}

	/**
	 * Retrieves detailed information for a particular VistaLink connector.
	 * @param serverName name of server to retrieve info for
	 * @param mbeanName canonical mbean Object Name of the ConnectorComponentRuntimeMBean to retrieve info for
	 * @return connectorInfoVO populated with connector information for the specified connector
	 * @va.exclude
	 */
	public static ConnectorInfoVO getConnectorDetail(String serverName, String mbeanName) throws Exception {

		ConnectorInfoVO connectorVO = null;

		try {
			MBeanHome remoteHome = ConsoleHelper.getRemoteHome(serverName);
			RemoteMBeanServer remoteMBeanServer = remoteHome.getMBeanServer();
			String domainName = remoteHome.getDomainName();

			Object[] params = new Object[] { mbeanName };
			String[] signature = new String[] { "java.lang.String" };

			String canonicalVljMcfMonitorName = VistaLinkRuntimeInfo.getServerSpecificCanonicalObjectName(domainName,
					serverName);

			if (logger.isDebugEnabled()) {
				logger.debug("about to make remote JMX retrieveConnectorDetailInfo call, to '"
						+ canonicalVljMcfMonitorName + "'.");
			}
			connectorVO = (ConnectorInfoVO) remoteMBeanServer.invoke(new ObjectName(canonicalVljMcfMonitorName),
					"retrieveConnectorDetailInfo", params, signature);

		} catch (MBeanHomeNotFoundException e) {
			// don't log
			throw e;
		} catch (InstanceNotFoundException e) {
			logger.warn(e);
			// could not find the 1-per-server VistaLinkRuntimeInfoMBean
			throw new VistaLinkMBeanNotFoundException(e);
		} catch (Exception e) {
			logger.error(e);
			throw e;
		}

		return connectorVO;
	}

	/**
	 * Utility function to return remote home.
	 * @param serverName server to return MBean home for
	 * @return MBeanHome for specified server
	 * @throws NamingException
	 * @throws MBeanHomeNotFoundException
	 */
	private static MBeanHome getRemoteHome(String serverName) throws NamingException, MBeanHomeNotFoundException {
		MBeanHome home = null;
		InitialContext ctx = new InitialContext();
		try {
			// get remote home
			home = (MBeanHome) ctx.lookup(MBeanHome.JNDI_NAME + "." + serverName);
		} catch (NamingException e) {
			String errorMessage = "Could not locate MBean Home for server '" + serverName
					+ "'. Check to see if this server has been started.";
			throw new MBeanHomeNotFoundException(errorMessage, e);
		}
		return home;
	}

	/**
	 * Compares two JNDI names.
	 * @param jndiName1
	 * @param jndiName2
	 * @return Returns true if both are a) non-null AND b) don't match
	 * @va.exclude
	 */
	public static boolean isJndiMismatch(String jndiName1, String jndiName2) {
		boolean returnVal = true;
		// returns true if either one is null -- something is wrong.
		if ((jndiName1 != null) && (jndiName2 != null)) {
			returnVal = (!(jndiName1.trim()).equals(jndiName2.trim()));
		}
		return returnVal;
	}

	public static String replaceBrWithLineBreak(String originalStr) {

		String returnVal = null;
		if (originalStr != null) {
			String[] splitStr = originalStr.split("<BR>");
			StringBuffer sb = new StringBuffer();
			for (int i = 0; i < splitStr.length; i++) {
				sb.append(splitStr[i]);
				sb.append("\n");
			}
			returnVal = sb.toString();
		}
		return returnVal;
	}

	/**
	 * Helper function for JSP pages to print out current time.
	 * @return String representation of current time.
	 * @va.exclude
	 */
	public static String getTimeStringNow() {
		SimpleDateFormat sdf = new SimpleDateFormat("MM/dd h:mm:ss a");
		return sdf.format(new Date());
	}
}