package gov.va.med.vistalink.adapter.cci;

import gov.va.med.vistalink.adapter.spi.ConnectorInfoVO;
import gov.va.med.vistalink.adapter.spi.EMAdapterEnvironment;
import gov.va.med.vistalink.adapter.spi.VistaLinkConnectionManager;
import gov.va.med.vistalink.adapter.spi.VistaLinkConnectionRequestInfo;
import gov.va.med.vistalink.adapter.spi.VistaLinkManagedConnectionFactory;
import gov.va.med.environment.Environment;
import gov.va.med.environment.ServerType;
import gov.va.med.exception.ExceptionUtils;

import java.io.PrintWriter;
import java.io.Serializable;

import javax.naming.Reference;
import javax.resource.NotSupportedException;
import javax.resource.Referenceable;
import javax.resource.ResourceException;
import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionFactory;
import javax.resource.cci.ConnectionSpec;
import javax.resource.cci.RecordFactory;
import javax.resource.cci.ResourceAdapterMetaData;
import javax.resource.spi.ConnectionManager;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.ManagedConnectionFactory;
import javax.resource.spi.ResourceAdapterInternalException;

import org.apache.log4j.Logger;
import org.apache.log4j.Priority;

/**
 * This implementation class provides an interface for getting connection to an EIS instance. For each type of adapter
 * derived from this class, it should overide/implement getConnection() to achieve adapter-specific connection.
 * 
 * @author Application Modernization - Foundations Team
 * @version 1.5.0.026
 */
public class VistaLinkConnectionFactory implements ConnectionFactory, Serializable, Referenceable {

	/**
	 * The logger used by this class
	 */
	private static final Logger logger = Logger.getLogger(VistaLinkConnectionFactory.class);

	/**
	 * strategy implementation to do JNDI mismatch checking
	 */
	private static JndiVerificationStrategy jndiVerificationStrategy = null;

	/**
	 * The managed connection factory used to allocate managed connections
	 */
	private ManagedConnectionFactory managedFactory;

	/**
	 * The connection manager used to allocate connection handles
	 */
	private ConnectionManager connectionManager;

	/**
	 * The reference
	 */
	private Reference reference;

	/**
	 * The PrintWriter used for logging to app server
	 */
	private PrintWriter printWriter;

	/**
	 * Stores result of per-connection JNDI mismatch checking
	 */
	private boolean passedJndiMismatchCheck = false;

	/**
	 * keeps track of the number of invocations of the J2SE getConnection method in a J2EE environment.
	 */
	private int getConnectionWrongModeInvocationCount = 0;

	private EMAdapterEnvironment adapterEnvironment;

	static {
		/**
		 * Set jndi verifications strategy based on server type
		 */
		if (Environment.getServerType().equals(ServerType.WEBLOGIC)) {
			jndiVerificationStrategy = new JndiVerificationStrategyWeblogic81();
		} else {
			jndiVerificationStrategy = new JndiVerificationStrategyAlwaysPass();
		}
	}

	/**
	 * @param managedFactory
	 *            the factory that the connection manager will use to construct managed connections
	 * @param connectionManager
	 *            the connection manager that will allocate new VistaLinkConnections. This value is null in an unmanaged
	 *            environment and defaults to VistaLinkConnectionManager this value is specified by the application
	 *            server in a managed environment
	 * @va.exclude
	 */
	public VistaLinkConnectionFactory(VistaLinkManagedConnectionFactory managedFactory,
			ConnectionManager connectionManager) {
		this.managedFactory = managedFactory;
		this.adapterEnvironment = managedFactory.getAdapterEnvironment();
		logger.debug("VistaLinkConnectionFactory constructor called!");

		if (connectionManager == null) {
			this.connectionManager = new VistaLinkConnectionManager();
		} else {
			this.connectionManager = connectionManager;
		}
		reference = null;
		printWriter = null;
	}

	/**
	 * Allocates a VistaLinkConnection Connection Handle using the specified connection manager
	 * 
	 * @return Connection The connection handle that may be used to interact with M.
	 * @throws ResourceException
	 */
	private Connection allocateConnection() throws ResourceException {
		if (managedFactory == null) {
			throw new ResourceAdapterInternalException("managedConnectionFactory is not set for this connectionFactory");
		}
		return (Connection) connectionManager.allocateConnection(managedFactory, null);
	}

	/**
	 * Allocates a VistaLinkConnection Connection Handle using the specified connection manager.
	 * 
	 * @return Connection The connection handle that may be used to interact with M.
	 * @throws ResourceException
	 */
	private Connection allocateConnection(ConnectionRequestInfo cri) throws ResourceException {
		if (managedFactory == null) {
			throw new ResourceAdapterInternalException("managedConnectionFactory is not set for this connectionFactory");
		}

		return (Connection) connectionManager.allocateConnection(managedFactory, cri);
	}

	/**
	 * @return PrintWriter
	 * @throws ResourceException
	 */
	public PrintWriter getLogWriter() throws ResourceException {
		return printWriter;
	}

	/**
	 * @param printWriter
	 * @throws ResourceException
	 * @va.exclude
	 */
	public void setLogWriter(PrintWriter printWriter) throws ResourceException {
		this.printWriter = printWriter;
	}

	/**
	 * Returns the meta data information associated with this adapter.
	 * 
	 * @see javax.resource.cci.ConnectionFactory#getMetaData()
	 * @va.exclude
	 */
	public ResourceAdapterMetaData getMetaData() throws ResourceException {
		return new VistaLinkResourceAdapterMetaData();
	}

	/**
	 * Referenceable interface.
	 * 
	 * @see javax.naming.Referenceable#getReference()
	 * @va.exclude
	 */
	public Reference getReference() {
		return reference;
	}

	/**
	 * Referenceable interface.
	 * 
	 * @see javax.resource.Referenceable#setReference(javax.naming.Reference)
	 * @va.exclude
	 */
	public void setReference(Reference reference) {
		this.reference = reference;
	}

	/**
	 * Gets a VistaLinkConnection handle with the connectionspec Empty method.
	 * 
	 * @throws NotSupportedException
	 * @see javax.resource.cci.ConnectionFactory#getConnection(javax.resource.cci.ConnectionSpec)
	 */
	public Connection getConnection(ConnectionSpec connectionSpec) throws ResourceException {

		try {
			if (adapterEnvironment.equals(EMAdapterEnvironment.J2SE)) {
				throw new VistaLinkResourceException(
						"Use of this method is not allowed in a J2SE environment, please use getConnection() instead");
			}

			// Reject JNDI mismatches (connectorConfig.xml JNDI vs. ra.xml JDNI)
			if (adapterEnvironment.equals(EMAdapterEnvironment.J2EE) && (!isPassedJndiMismatchCheck())) {
				if (!jndiVerificationStrategy.checkJndiMismatch(
						(VistaLinkManagedConnectionFactory) this.managedFactory, this)) {
					//TODO throw new type of error -- VistaLinkConfigurationException?
					String errorMessage = "JNDI Mismatch found for connector configured with JNDI Name: '"
							+ ((this.managedFactory == null) ? "[not defined]"
									: ((VistaLinkManagedConnectionFactory) this.managedFactory).getConnectorJndiName())
							+ "'. Rejecting connection request.";
					logger.error(errorMessage);
					throw new VistaLinkResourceException(errorMessage);
				} else {
					setPassedJndiMismatchCheck(true);
				}
			}
			return allocateConnection(new VistaLinkConnectionRequestInfo(connectionSpec));

		} catch (ResourceException e) {

			if (logger.isEnabledFor(Priority.ERROR)) {

				String errMsg = (new StringBuffer()).append("Can not allocate VistaLinkConnection.").append("\n\t")
						.append(ExceptionUtils.getFullStackTrace(e)).toString();

				logger.error(errMsg);
			}

			throw e;
		}
	}

	/**
	 * Gets a VistaLinkConnection connection handle; calls allocateConnection()
	 * 
	 * @see javax.resource.cci.ConnectionFactory#getConnection()
	 */
	public Connection getConnection() throws ResourceException {

		if (adapterEnvironment.equals(EMAdapterEnvironment.J2EE)) {

			//TODO: revisit periodically to see if WebLogic still calls this
			// method once per deployment.
			String errMsg = "Use of this method is not allowed in a J2EE environment, please use getConnection(ConnectionSpec) instead. NOTE: IF this is exception is logged during connector deployment or startup on BEA WebLogic, one instance of this exception per connector module is always thrown. This is normal and can be ignored in those cases.";
			this.getConnectionWrongModeInvocationCount++;
			if ((Environment.getServerType().equals(ServerType.WEBLOGIC))
					&& (this.getConnectionWrongModeInvocationCount < 2)) {
				logger.debug(errMsg);
			} else {
				logger.error(errMsg);
			}
			throw new VistaLinkResourceException(errMsg);
		}

		try {

			return allocateConnection();

		} catch (ResourceException e) {

			if (logger.isEnabledFor(Priority.ERROR)) {

				String errMsg = (new StringBuffer()).append("Can not allocate VistaLinkConnection.").append("\n\t")
						.append(ExceptionUtils.getFullStackTrace(e)).toString();

				logger.error(errMsg);
			}

			throw e;
		}
	}

	/**
	 * Empty method.
	 * 
	 * @throws NotSupportedException
	 * @see javax.resource.cci.ConnectionFactory#getRecordFactory()
	 * @va.exclude
	 */
	public RecordFactory getRecordFactory() throws ResourceException {
		throw new NotSupportedException("getRecordFactory() is not supported at this time.");
	}

	/**
	 * Creates VistaLinkConnectionFactory to be used in non-managed environment (J2SE).
	 * <p>
	 * This method should never be called in managed environment (J2EE) as application server is responsible for
	 * creating ConnectionFactories.
	 * <p>
	 * Client code in that case uses JNDI lookup to get a reference to the ConnectionFactory.
	 * 
	 * @param hostIPAddress -
	 *            the IP address to connect to
	 * @param hostPort -
	 *            the port to open
	 * @return VistaLinkConnectionFactory The connection factory that will return Connections for the specified
	 *         IPAddress and port
	 * @throws ResourceException
	 */

	public static final VistaLinkConnectionFactory getVistaLinkConnectionFactory(String hostIPAddress, Integer hostPort)
			throws ResourceException {
		try {
			VistaLinkManagedConnectionFactory mcf = new VistaLinkManagedConnectionFactory();
			mcf.setNonManagedHostIPAddress(hostIPAddress);
			mcf.setNonManagedHostPort(hostPort.intValue());
			mcf.setAdapterEnvironment(EMAdapterEnvironment.J2SE);

			return (VistaLinkConnectionFactory) mcf.createConnectionFactory();
		} catch (ResourceException e) {
			if (logger.isEnabledFor(Priority.ERROR)) {

				String errMsg = (new StringBuffer()).append("Can not create VistaLinkConnectionFactory.")
						.append("\n\t").append(ExceptionUtils.getFullStackTrace(e)).toString();
				logger.error(errMsg);
			}
			throw e;
		}
	}

	/**
	 * @return Returns the passedJndiMismatchCheck.
	 */
	private synchronized boolean isPassedJndiMismatchCheck() {
		return passedJndiMismatchCheck;
	}

	/**
	 * @param passedJndiMismatchCheck
	 *            The passedJndiMismatchCheck to set.
	 */
	private synchronized void setPassedJndiMismatchCheck(boolean passedJndiMismatchCheck) {
		this.passedJndiMismatchCheck = passedJndiMismatchCheck;
	}

	/**
	 * Populate a ConnectorInfoVO object with information from the stored VistaLinkManagedConnectionFactory instance.
	 * @param connectorInfoVO
	 * @va.exclude
	 */
	public void getMCFInfo(ConnectorInfoVO connectorInfoVO, boolean getDetailInfo) {
		if (this.managedFactory != null) {
			VistaLinkManagedConnectionFactory mcf = (VistaLinkManagedConnectionFactory) this.managedFactory;

			connectorInfoVO.setCfgIpAddress(mcf.getHostIpAddressUnresolved());
			connectorInfoVO.setCfgPort(mcf.getHostPort());

			connectorInfoVO.setHlthConnectionAuthFailureCount(mcf.getConnectionAuthFailureCount());
			connectorInfoVO.setHlthConnectionFailureCount(mcf.getConnectionFailureCount());
			connectorInfoVO.setHlthDivisionMismatchCount(mcf.getDivisionMismatchCount());
			connectorInfoVO.setHlthIdentityFailureCount(mcf.getIdentityFailureCount());
			connectorInfoVO.setHlthProductionMismatchCount(mcf.getProductionMismatchCount());
			connectorInfoVO.setCfgTimeout((new Long(mcf.getDefaultJ2EETimeOut() / 1000)).intValue());
			connectorInfoVO.setCfgTimeoutAlwaysUseDefaultAsMin(mcf.isAlwaysUseDefaultAsMin());

			if (getDetailInfo) {
				connectorInfoVO.setPerfCreateConnectionHandleAvgMillis(mcf.getCreateConnectionHandleAvgMillis());
				connectorInfoVO.setPerfMatchManagedConnectionAvgMillis(mcf.getMatchManagedConnectionAvgMillis());
				connectorInfoVO.setVljDistinguishedIdentifier(mcf.getDistinguishedIdentifier());
				connectorInfoVO.setMSystemInfo(mcf.getSystemInfo());
			}
		}
	}
}