package gov.va.med.vistalink.adapter.spi;

import gov.va.med.crypto.DESPassPhraseEncrypter;
import gov.va.med.crypto.EncrypterRegistry;
import gov.va.med.vistalink.adapter.cci.VistaLinkResourceException;
import gov.va.med.vistalink.adapter.config.Connector;
import gov.va.med.vistalink.adapter.config.ConnectorsDocument;
import gov.va.med.vistalink.institution.InstitutionMappingBadStationNumberException;
import gov.va.med.vistalink.institution.InstitutionMappingImpl;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;

import org.apache.log4j.Logger;
import org.apache.xmlbeans.XmlException;

/**
 * Reads in managed connection factory configuration from vistalink configuration file. Isolates J2SE mode's use of MCF
 * from any XML Beans dependencies as a result of J2EE mode's use of XML Beans.
 * 
 * @author Application Modernization - Foundations Team
 * @version 1.5.0.026
 */
public class ConfigurationReader {

  public static final String CONFIG_FILE_PATH = "/gov.va.med.vistalink.connectorConfig.xml";

  /**
   * Parses config file using XMLBeans-generated config classes.
   * 
   * @param mcf ManagedConnectionFactory to retrieve/set properties for
   * @throws ResourceException
   */
  protected static void parseConfig(VistaLinkManagedConnectionFactory mcf) throws VistaLinkResourceException {

    Logger logger = Logger.getLogger(ConfigurationReader.class);

    //TODO unit tests for parseConfig
    String connectorJndiName = mcf.getConnectorJndiName();
    boolean saveFile = false;

    if (logger.isDebugEnabled()) {
      logger.debug("ConnectorJndiName index: " + connectorJndiName);
    }

    if (connectorJndiName == null) {
      String error = "Connector descriptor is not configured with connectorJndiName to be retrieved from configuration file.";
      throw new VistaLinkResourceException(error);
    }

    InputStream connectorConfig = null;
    int enabledMatchCount = 0;

    try {

      try {
        connectorConfig = ConfigurationReader.class.getResourceAsStream(CONFIG_FILE_PATH);
        logger.info("Opened VistaLink configuration file at: " + ConfigurationReader.class.getResource(ConfigurationReader.CONFIG_FILE_PATH).getPath());
      } catch (Exception e) {
        throw new VistaLinkResourceException("Could not open vistalink configuration file '" + CONFIG_FILE_PATH
            + "'.", e);
      }

      ConnectorsDocument config = ConnectorsDocument.Factory.parse(connectorConfig);
      Connector[] connectors = config.getConnectors().getConnectorArray();
      for (int i = 0; i < connectors.length; i++) {
        if (logger.isDebugEnabled()) {
          logger.debug("Given connectorJndiName '" + connectorJndiName + "' from ra.xml custom property");
        }

        if (!connectors[i].getEncrypted()) {
          long start = 0;
          if (logger.isDebugEnabled()) {
            logger.debug("found unencrypted entries, encrypting for: " + connectors[i].getJndiName());
            start = System.currentTimeMillis();
          }
          connectors[i].setAccessCode(getEncrypted(connectors[i].getAccessCode()));
          connectors[i].setVerifyCode(getEncrypted(connectors[i].getVerifyCode()));
          connectors[i].setEncrypted(true);
          saveFile = true;
          if (logger.isDebugEnabled())
            logger.debug("finished encrypting (" + (System.currentTimeMillis() - start)
                + " ms) entries for: " + connectors[i].getJndiName());
        }

        if (connectorJndiName.equals(connectors[i].getJndiName())) {

          if (connectors[i].getEnabled()) {

            enabledMatchCount++;
            if (enabledMatchCount > 1) {
              throw new VistaLinkResourceException(
                  "Multiple entries in VistaLink configuration file are enabled and have the same connectorJndiName, '"
                      + connectorJndiName + "'.");
            }

            logger.debug("found match.");
            mcf.setHostIpAddress(connectors[i].getIp());
            mcf.setHostPort(connectors[i].getPort());

            if (connectors[i].getEncrypted()) {
              logger.debug("found encrypted entries, decrypting...");
              mcf.setAccessCode(getDecrypted(connectors[i].getAccessCode()));
              mcf.setVerifyCode(getDecrypted(connectors[i].getVerifyCode()));
            } else
              logger.warn("unencrypted entry in the config file! - " + connectorJndiName);

            mcf.setDefaultJ2EETimeOut(connectors[i].getTimeout() * 1000);
            mcf.setAlwaysUseDefaultAsMin(connectors[i].getAlwaysUseDefaultAsMin());
            String primaryStation = Integer.toString(connectors[i].getPrimaryStation());
            // append suffix if present
            if (connectors[i].getPrimaryStationSuffix() != null) {
              primaryStation = primaryStation + connectors[i].getPrimaryStationSuffix();
            }
            mcf.setPrimaryStation(primaryStation);

            // load/reload station #s for this connector
            if (primaryStation.length() > 0) {
              InstitutionMappingImpl instMap = InstitutionMappingImpl.getInstance();
              String[] stationNumbersToMap = new String[] { primaryStation };
              try {
                instMap.loadMappingsForJndiName(connectorJndiName, stationNumbersToMap, mcf
                    .getDefaultJ2EETimeOut());
              } catch (InstitutionMappingBadStationNumberException e1) {
                throw new VistaLinkResourceException(
                    "Could not load institution mapping for connector JNDI name: '"
                        + connectorJndiName + "' due to bad station number(s): '"
                        + e1.getMessage() + "'.", e1);
              }
            } else {
              throw new VistaLinkResourceException(
                  "primary Station not found; this attribute is required for connector configuration.");
            }

            // do this last -- only enable if we successfully parse the entire config file
            mcf.setConnectorEnabled(connectors[i].getEnabled());

          } else {

            if (logger.isDebugEnabled()) {
              logger.debug("Skipping disabled connector '" + connectorJndiName + "'.");
            }
          }
        }
        // if no match is found, connectorEnabled will be false, and
        // connector will not be enabled.

        //run thru the rest of the file and encrypr the entries

      }
      if (saveFile)
        config.save(new File(ConfigurationReader.class.getResource(ConfigurationReader.CONFIG_FILE_PATH)
            .getPath()));
    } catch (VistaLinkResourceException e) {
      // already created/logged inside
      throw e;
    } catch (XmlException e) {
      String error = "Problem parsing config file. ";
      throw new VistaLinkResourceException(error, e);
    } catch (IOException e) {
      String error = "Problem opening or saving a config file. ";
      throw new VistaLinkResourceException(error, e);
    } catch (Exception e) {
      // catches XML Beans exceptions
      String error = "Problem parsing config file. ";
      throw new VistaLinkResourceException(error, e);
    } finally {
      if (connectorConfig != null) {
        try {
          connectorConfig.close();
        } catch (IOException e1) {
          logger.error("Problem closing config file input stream: ", e1);
        }
      }
    }

    if (enabledMatchCount == 0) {
      throw new VistaLinkResourceException("Could not find matching configuration for connectorJndiName '"
          + connectorJndiName + "'.");
    }
  }

  private static String getDecrypted(String encrypted) {
    return EncrypterRegistry.getInstance(DESPassPhraseEncrypter.class).decrypt(encrypted);
  }

  private static String getEncrypted(String original) {
    return EncrypterRegistry.getInstance(DESPassPhraseEncrypter.class).encrypt(original);
  }
}