package gov.va.med.vistalink.adapter.spi;

import gov.va.med.environment.Environment;
import gov.va.med.environment.ServerType;
import gov.va.med.exception.ExceptionUtils;
import gov.va.med.vistalink.adapter.cci.VistaLinkConnectionFactory;
import gov.va.med.vistalink.adapter.cci.VistaLinkResourceException;
import gov.va.med.vistalink.adapter.heartbeat.VistaHeartBeatTimerManager;
import gov.va.med.vistalink.institution.InstitutionMapManager;
import gov.va.med.vistalink.institution.InstitutionMappingImpl;

import java.io.PrintWriter;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.resource.ResourceException;
import javax.resource.spi.ConnectionManager;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.ManagedConnection;
import javax.resource.spi.ManagedConnectionFactory;
import javax.security.auth.Subject;

import org.apache.log4j.Logger;
import org.apache.log4j.Priority;

/**
 * An object of this class is a factory of both ManagedConnection and connection factory instances. This class supports
 * connection pooling by defining methods for matching and creating connections.
 * 
 * @author Application Modernization - Foundations Team
 * @version 1.5.0.026
 */
public class VistaLinkManagedConnectionFactory implements ManagedConnectionFactory, Serializable {

	public static final String ADAPTER_VERSION = "1.5";

	public static final String ADAPTER_FULL_VERSION = "1.5.0.026";

	public static final String CONNECTOR_PROPERTY_NAME = "connectorJndiName";

	private static final int HEARTBEAT_RATE_MULTIPLIER = 2;

	/**
	 * Cache the classes' hashcode
	 */
	private int hashCode;
	
	/**
	 * constant for socket timeout used for J2EE
	 */
	private static final long J2EE_DEFAULT_SOCKET_TIMEOUT = 5000;

	/**
	 * The logger for this class
	 */
	private static final Logger logger = Logger.getLogger(VistaLinkManagedConnectionFactory.class);

	/**
	 * The minimum value allowed for the timeout
	 */
	private long defaultJ2EETimeOut = J2EE_DEFAULT_SOCKET_TIMEOUT;

	/**
	 * A flag wheter to use the default minimum value or not
	 */
	private boolean alwaysUseDefaultAsMin = true;

	/**
	 * The raw (unresolved) ip address of the Host M Server this factory will create connections to.
	 */
	private String hostIpAddress;

	/**
	 * The primaryStation attribute assigned from the config file for institution mapping purposes.
	 */
	private String primaryStation = "";

	/**
	 * The primaryStationSuffix attribute assigned from the config file for institution mapping purposes, for the
	 * special case of AAC station #s e.g. 200M.
	 */
	private String primaryStationSuffix = "";

	/**
	 * The resolved ip address of the Host M Server this factory will create connections to.
	 */
	private InetAddress hostIpAddressResolved = null;

	/**
	 * The port of the Host M Server this factory will create connections to
	 */
	private int hostPort;

	/**
	 * The access code used to perform initial authentication
	 */
	private String accessCode;

	/**
	 * The verify code used to perform initial authentication
	 */
	private String verifyCode;

	/**
	 * The connector name index to retrieve the connector params from the config file
	 */
	private String connectorJndiName = "";

	/**
	 * retains connection factory created by CreateConnectionFactory()
	 */
	private VistaLinkConnectionFactory cf;

	/**
	 * connector enabled or disabled -- config file setting
	 */
	private boolean connectorEnabled = false;

	/**
	 * The maximum number of connection handles that this factory's managed connections will be able to allocate
	 */
	private int maxConnectionHandles = 1;

	/**
	 * The log writer that will be set by the application server
	 */
	private PrintWriter printWriter;

	/**
	 * The socket time out multiplier. This value is multiplied by the heart beat rate to determine the socket time out
	 */
	private int heartBeatRateMultiplierForSocketTimeOut;

	/**
	 * The timer manager for this factory and its created managed connections
	 */
	private VistaHeartBeatTimerManager timerManager;

	/**
	 * Identifies the environment that this adapter is running in
	 */
	private EMAdapterEnvironment adapterEnvironment = EMAdapterEnvironment.J2SE;

	private long managedConnectionCount = 0;

	private Object mcCountSyncObj = new Object();

	private static long managedFactoryCount = 0;

	private static Object mfCountSyncObj = new Object();

	private long distinguishedIdentifier = 0;

	// health monitoring fields
	private long connectionFailureCount = 0;

	private long connectionAuthFailureCount = 0;

	private long divisionMismatchCount = 0;

	private long identityFailureCount = 0;

	private long productionMismatchCount = 0;

	private Object healthSyncObj = new Object();

	// perf stat fields
	private long createConnectionHandleCount = 0;

	private long createConnectionHandleTotalMillis = 0;

	private Object createConnectionHandleAvgSyncObj = new Object();

	private long matchManagedConnectionCount = 0;

	private long matchManagedConnectionTotalMillis = 0;

	private Object matchManagedConnectionAvgSyncObj = new Object();

	private String parseConfigFailureMessage = null;
	
	private HashSet mcSet = new HashSet();

	// static initializer
	static {

		logger.info("Creating managed connection factory, VistALink adapter version " + ADAPTER_FULL_VERSION + ".");
		
		ServerType serverType = Environment.getServerType();
		
		if (serverType.equals(ServerType.WEBLOGIC)) {
			// create VistaLinkRuntimeInfo
			VistaLinkRuntimeInfo.register();
		} else if (serverType.equals(ServerType.UNKNOWN)) {
			logger
					.warn("gov.va.med.environment.servertype is not defined for this JVM. For J2EE systems only -- check this -D JVM arg; MBeans required for the VistaLink console will not be loaded.");
		} else if (serverType.equals(ServerType.JAVA_SE)) {
			// no problem, nothing to log -- we skip for Java SE anyway
		} else {
			logger.warn("MBeans required for the VistaLink console are not supported for server type "
					+ serverType + " and will not be loaded.");
		}

		// create/initialize institution mapping singleton
		if (!InstitutionMappingImpl.instanceExists()) {
			try {
				// if doesn't exist, create
				// note: if this is j2se, not a problem -- it just creates empty object.
				logger.debug("Creating institution mapping singleton instance.");
				InstitutionMappingImpl.createInstance();
				// register MBean. Note: uses WLS-specific MBean stuff
				if (serverType.equals(ServerType.WEBLOGIC)) {
					InstitutionMapManager.register();
				}
			} catch (VistaLinkResourceException e) {
				logger.error("Could create institution mapping singleton instance.", e);
			} catch (Exception e) {
				logger.error("Could create institution mapping singleton instance.", e);
			}
		}
	}

	/**
	 * Constructor. Actions include creating a new VistaHeartBeatTimerManager, setting the adapter environment to J2EE,
	 * setting the instance distinguished identifier and incrementing the class distinguished identifier, and
	 * registering the VistaLink MBean for this managed connection factory.
	 */
	public VistaLinkManagedConnectionFactory() {

		// set distinguished id, hashcode first
		synchronized (mfCountSyncObj) {
			distinguishedIdentifier = ++managedFactoryCount;
		}
		setHashCode();

		hostIpAddressResolved = null;
		timerManager = new VistaHeartBeatTimerManager();
		timerManager.setAdapterEnvironment(adapterEnvironment);
		setHeartBeatRateMultiplierForSocketTimeOut(HEARTBEAT_RATE_MULTIPLIER);

		// NOTE: ra.xml setter/getter properties are *not* yet set here, in WLS
		// at least

		if (logger.isDebugEnabled()) {
			String debugMsg = getLoggerFormattedString("Constructed");
			logger.debug(debugMsg);
		}
	}

	/**
	 * Creates a connection factory instance. The connection factory instance gets initialized with a default connection
	 * manager provided by the resource adapter.
	 * 
	 * @return VistaLinkConnectionFactory instance
	 * @see javax.resource.spi.ManagedConnectionFactory#createConnectionFactory(javax.resource.spi.ConnectionManager)
	 */
	public Object createConnectionFactory(ConnectionManager mgr) throws ResourceException {

		// NOTE: ra.xml setter/getter properties *are* set when get here, in WLS
		// at least

		if (logger.isDebugEnabled()) {
			String loggerMsg = getLoggerFormattedString("createConnectionFactory->managed");
			logger.debug(loggerMsg);
		}

		// TODO: use NDS service to retrieve IP address and port based
		// on station #.

		cf = new VistaLinkConnectionFactory(this, mgr);

		return cf;
	}

	/**
	 * Creates a connection factory instance with the default connection manager.
	 * 
	 * @see javax.resource.spi.ManagedConnectionFactory#createConnectionFactory()
	 */
	public Object createConnectionFactory() throws ResourceException {

		if (logger.isDebugEnabled()) {
			String loggerMsg = getLoggerFormattedString("createConnectionFactory->unmanaged");
			logger.debug(loggerMsg);
		}
		cf = new VistaLinkConnectionFactory(this, null);
		return cf;
	}

	public VistaLinkSystemInfoVO getSystemInfo() {
		VistaLinkManagedConnection mc = null;
		VistaLinkSystemInfoVO systemInfoVO = null;
		try {
			mc = (VistaLinkManagedConnection) createManagedConnection(null, null);
			systemInfoVO = mc.executeSystemInfoInteraction().getSystemInfoVO();
			mc.destroy();
		} catch (Exception e) {
			logger.error(e);
			systemInfoVO = new VistaLinkSystemInfoVO();
			systemInfoVO.setErrorMessage(e.toString());
		}
		return systemInfoVO;
	}

	/**
	 * Creates a new physical connection to M.
	 * 
	 * @param subject Caller's security information
	 * @param info Additional resource adapter specific connection request information
	 * @return ManagedConnection instance
	 * @throws VistaLinkResourceException
	 * @see javax.resource.spi.ManagedConnectionFactory#createManagedConnection(javax.security.auth.Subject,
	 *      javax.resource.spi.ConnectionRequestInfo)
	 */
	public ManagedConnection createManagedConnection(Subject subject, ConnectionRequestInfo info)
			throws ResourceException {

		try {

			// if disabled, stop here
			if (!this.connectorEnabled) {
				StringBuffer failureMessage = new StringBuffer("Connector '");
				failureMessage.append(this.connectorJndiName);
				failureMessage.append("' is disabled; check vistalink config file. Not returning connection.");
				if (parseConfigFailureMessage != null) {
					failureMessage.append(" Deploy-time failure was: '");
					failureMessage.append(this.parseConfigFailureMessage);
					failureMessage.append("'; fix config file and redeploy connector.");
				}
				throw new VistaLinkResourceException(failureMessage.toString());
			}

			// resolve IP address (in case it needs name resolution) and set
			// internal hostIpAddressResolved property
			setHostIpAddressResolved(this.hostIpAddress);

			// now ready to actually create the connection
			long mcDistinguishedIdentifier = 0;
			synchronized (mcCountSyncObj) {
				mcDistinguishedIdentifier = ++managedConnectionCount;
			}
			if (logger.isDebugEnabled()) {
				String loggerMsg = getLoggerFormattedString((new StringBuffer())
						.append("Creating Managed connection->").append(mcDistinguishedIdentifier).toString());
				logger.debug(loggerMsg);
			}

			// create the connection
			VistaLinkManagedConnection mc = new VistaLinkManagedConnection(this, mcDistinguishedIdentifier);
			if (logger.isDebugEnabled()) {
				String loggerMsg = getLoggerFormattedString((new StringBuffer()).append("Created Managed connection->")
						.append(mcDistinguishedIdentifier).toString());
				logger.debug(loggerMsg);
			}
			mcSet.add(mc);
			return mc;

		} catch (UnknownHostException e) {
			String errMsg = "Unknown Host";
			if (logger.isEnabledFor(Priority.ERROR)) {
				String loggerMsg = getLoggerFormattedStringWStackTrace(errMsg, e);
				logger.error(loggerMsg);
			}
			throw new VistaLinkResourceException(errMsg, e);
		} catch (VistaLinkResourceException e) {
			if (logger.isEnabledFor(Priority.ERROR)) {
				String loggerMsg = getLoggerFormattedStringWStackTrace(
						"Resource exception occurred in creatingManagedConnection()", e);
				logger.error(loggerMsg);
			}
			throw e;
		} catch (Exception e) {
			if (logger.isEnabledFor(Priority.ERROR)) {
				String loggerMsg = getLoggerFormattedStringWStackTrace(
						"Unexpected exception occured in creatingManagedConnection()", e);
				logger.error(loggerMsg);
			}
			throw new VistaLinkResourceException(e);
		}
	}

	/**
	 * Remove managed connection for the set used for lookup in matchedManagedConnections
	 * 
	 * @param mc reference to a managed connection
	 */
	protected void removeMc(VistaLinkManagedConnection mc) {
		mcSet.remove(mc);
	}

	/**
	 * Returns a matched connection from the candidate set of connections. This method returns a ManagedConnection
	 * instance that is the best match for handling the connection allocation request.
	 * <p>
	 * The method iterates through a Set of type VistaLinkManagedConnection determines if the HostIPAddress and HostPort
	 * are the same as those defined for this instance of the VistaLinkManagedConnectionFactory and then determines if
	 * the connection request info is equal to the connection request info associated with the mc.
	 * <p>
	 * <br>
	 * <b>Managed Connection Check Criteria: </b> <br>
	 * 1) If the connection request info passed in is equal to the connection request info associated with the mc, then
	 * this mc is immediately returned. It is the best match.
	 * <p>
	 * 2) The first mc found that is not already associated with a connection request info object (mc.getConnReqInfo() ==
	 * null), is the second best match. This mc is returned if no match for #1 is found.
	 * <p>
	 * 3) The last mc found where the connection request info passed in is not equal to the connection request info
	 * associated with the mc will be returned if no match for either #1 or #2 is found.
	 * <p>
	 * 4) If no match for any of #1, #2, #3 is found then null is returned.
	 * <p>
	 * 
	 * @param connectionSet candidate connection set
	 * @param subject caller's security information
	 * @param info additional resource adapter specific connection request information
	 * @return ManagedConnection if resource adapter finds an acceptable match otherwise null
	 * @throws VistaLinkResourceException
	 * @see javax.resource.spi.ManagedConnectionFactory#matchManagedConnections(java.util.Set,
	 *      javax.security.auth.Subject, javax.resource.spi.ConnectionRequestInfo)
	 */
	public ManagedConnection matchManagedConnections(Set connectionSet, Subject subject, ConnectionRequestInfo info)
			throws ResourceException {
		long beginTimeMillis = System.currentTimeMillis();
		String loggerMsg;
		VistaLinkManagedConnection nullConnSpecConnection = null;
		VistaLinkManagedConnection lastConnection = null;
		VistaLinkManagedConnection returnConnection = null;
		boolean foundMatchWithOutConnSpec = false;

		if (logger.isDebugEnabled()) {
			loggerMsg = getLoggerFormattedString("matchManagedConnection: Pool Size: " + connectionSet.size());
			logger.debug(loggerMsg);
		}

		VistaLinkConnectionRequestInfo vlConReqInfo = (VistaLinkConnectionRequestInfo) info;
		if (info != null) {
			if (logger.isDebugEnabled()) {
				loggerMsg = getLoggerFormattedString((new StringBuffer()).append("matchManagedConnection->").append(
						info.toString()).toString());
				logger.debug(loggerMsg);
			}
		}

		Iterator it = connectionSet.iterator();
		while (it.hasNext()) {
			Object obj = it.next();
			if (obj instanceof VistaLinkManagedConnection) {
				VistaLinkManagedConnection mc = (VistaLinkManagedConnection) obj;
				if (logger.isDebugEnabled()) {
					loggerMsg = getLoggerFormattedString((new StringBuffer()).append("matchManagedConnection->")
							.append(mc.toString()).append("->").append(mc.getConnReqInfoString()).toString());
					logger.debug(loggerMsg);
				}
				if (mcSet.contains(mc)) {
					lastConnection = mc;
					//is this mc re-authenticated with the same
					// reqinfo/connspec
					if (vlConReqInfo.equals(mc.getConnReqInfo())) {
						if (logger.isDebugEnabled()) {
							loggerMsg = getLoggerFormattedString((new StringBuffer()).append(
									"matchManagedConnection->Found match with matching connection request info->")
									.append(mc.toString()).append("->").append(mc.getConnReqInfoString()).toString());
							logger.debug(loggerMsg);
						}
						// save elapsed time
						// TODO save conditionally based on config setting
						long elapsedMillis = System.currentTimeMillis() - beginTimeMillis;
						addMatchManagedConnectionTime(elapsedMillis);
						if (logger.isInfoEnabled()) {
							logger.info(getLoggerFormattedString(new StringBuffer().append(
									"matchManagedConnections(...) processing time [match found] == ").append(
									elapsedMillis).toString()));
						}
						return mc;

					} else if (!foundMatchWithOutConnSpec) {
						if (mc.getConnReqInfo() == null) {
							foundMatchWithOutConnSpec = true;
							nullConnSpecConnection = mc;
						}
						if (logger.isDebugEnabled()) {
							loggerMsg = getLoggerFormattedString((new StringBuffer()).append(
									"matchManagedConnection->no match on request info->").append(mc.toString()).append(
									"->").append(mc.getConnReqInfoString()).toString());
							logger.debug(loggerMsg);
						}
					} // !foundMatchWithOutConnSpec check
				} // mcSet check
			} // instanceof check
		} // while loop

		if (nullConnSpecConnection != null) {
			returnConnection = nullConnSpecConnection;
		} else {
			returnConnection = lastConnection;
		}
		// save end time
		// TODO save conditionally based on config setting
		long elapsedMillis = System.currentTimeMillis() - beginTimeMillis;
		addMatchManagedConnectionTime(elapsedMillis);
		if (logger.isInfoEnabled()) {
			logger.info(getLoggerFormattedString(new StringBuffer().append(
					"matchManagedConnections(...) processing time [match not found] == ").append(elapsedMillis)
					.toString()));
		}
		return returnConnection;
	}

	/**
	 * @see javax.resource.spi.ManagedConnectionFactory#getLogWriter()
	 */
	public PrintWriter getLogWriter() throws ResourceException {
		if (printWriter == null) {
			String errMsg = "Log writer is null";
			if (logger.isEnabledFor(Priority.ERROR)) {
				String loggerMsg = getLoggerFormattedString(errMsg);
				logger.error(loggerMsg);
			}
			throw new VistaLinkResourceException(errMsg);
		}
		return printWriter;
	}

	/**
	 * @see javax.resource.spi.ManagedConnectionFactory#setLogWriter(java.io.PrintWriter)
	 */
	public void setLogWriter(PrintWriter out) throws ResourceException {
		printWriter = out;
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {

		if (obj == this) {
			return true;
		} else if (!(obj instanceof VistaLinkManagedConnectionFactory)) {
			return false;
		} else {
			VistaLinkManagedConnectionFactory objMcf = (VistaLinkManagedConnectionFactory) obj;
			if (objMcf.distinguishedIdentifier == this.distinguishedIdentifier) {
				return true;
			} else {
				return false;
			}
		}
	}

	/**
	 * Sets the cached hashcode for this object.
	 */
	private void setHashCode() {
		
		// algorithm taken from "Effective Java" item #8.
		int HASHCODE_SEED = 17;
		int returnVal = HASHCODE_SEED;

		// distinguished identifier contribution to hashcode
		int distIdHashCode = (int) (this.distinguishedIdentifier ^ (this.distinguishedIdentifier >>> 32));
		returnVal = 37 * returnVal + distIdHashCode; 

		this.hashCode = returnVal;
	}

	/**
	 * return cached hashcode
	 * 
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return hashCode;
	}
	
	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		String adapterEnvironmentString = "";
		if (adapterEnvironment != null) {
			adapterEnvironmentString = adapterEnvironment.toString();
		}
		return ((new StringBuffer()).append(this.getClass().getName()).append("[]").append(hostIpAddress).append("[]")
				.append(hostPort).append("[]").append(maxConnectionHandles).append("[adapterversion]").append(
						ADAPTER_FULL_VERSION).append("[]").append(adapterEnvironmentString).append("[fdi]")
				.append(distinguishedIdentifier)).toString();

	}

	/**
	 * Returns the maxConnectionHandle.
	 * 
	 * @return int
	 */
	public int getMaxConnectionHandles() {
		return maxConnectionHandles;
	}

	/**
	 * Sets the maxConnectionHandle.
	 * 
	 * @param maxConnectionHandle The maxConnectionHandle to set
	 */
	private void setMaxConnectionHandles(int maxConnectionHandles) {
		this.maxConnectionHandles = maxConnectionHandles;
	}

	/**
	 * Returns the hostPort.
	 * 
	 * @return int
	 */
	public int getHostPort() {
		return hostPort;
	}

	/**
	 * used for non-managed mode only
	 * 
	 * Sets the hostPort.
	 * 
	 * @param hostPort The hostPort to set
	 */
	public void setNonManagedHostPort(int hostPort) {
		this.hostPort = hostPort;
		// additionally this must be a J2SE adapter
		setJ2SEProperties();
	}

	/**
	 * used for non-managed mode only
	 * 
	 * sets the host IP Address
	 * 
	 * @param hostipaddress
	 */
	public void setNonManagedHostIPAddress(String hostipaddress) {
		this.hostIpAddress = hostipaddress;
		// additionally this must be a J2SE adapter
		setJ2SEProperties();
	}

	/**
	 * Gets the host IP Address.
	 * 
	 * @return String
	 */
	private String getHostIPAddress() {
		return hostIpAddress;
	}

	/**
	 * 
	 * @param log
	 * @return String
	 */
	private String getLoggerFormattedString(String log) {
		return (new StringBuffer()).append(this.toString()).append("\n\t").append(log).toString();
	}

	/**
	 * 
	 * @param e
	 * @return String
	 */
	private String getLoggerFormattedStringWStackTrace(Throwable e) {

		return getLoggerFormattedString(ExceptionUtils.getFullStackTrace(e));

	}

	/**
	 * 
	 * @param log
	 * @param e
	 * @return String
	 */
	private String getLoggerFormattedStringWStackTrace(String log, Throwable e) {

		return getLoggerFormattedString((new StringBuffer()).append(log).append("\n\t").append(
				ExceptionUtils.getFullStackTrace(e)).toString());

	}

	/**
	 * Returns the heartBeatRateMultiplierForSocketTimeOut.
	 * 
	 * @return int
	 */
	public int getHeartBeatRateMultiplierForSocketTimeOut() {
		return heartBeatRateMultiplierForSocketTimeOut;
	}

	/**
	 * Sets the heartBeatRateMultiplierForSocketTimeOut.
	 * 
	 * @param heartBeatRateMultiplierForSocketTimeOut The heartBeatRateMultiplierForSocketTimeOut to set
	 */
	public void setHeartBeatRateMultiplierForSocketTimeOut(int heartBeatRateMultiplierForSocketTimeOut) {
		this.heartBeatRateMultiplierForSocketTimeOut = heartBeatRateMultiplierForSocketTimeOut;
	}

	/**
	 * gets the timeout value for managed connections
	 * 
	 * @return int
	 */
	protected long getSocketTimeOut() {
		if (this.adapterEnvironment.equals(EMAdapterEnvironment.J2SE)) {
			return getHeartBeatRateMultiplierForSocketTimeOut() * (int) timerManager.getHeartBeatRate();
		} else {
			return getDefaultJ2EETimeOut();
		}
	}

	/**
	 * Returns the timerManager.
	 * 
	 * @return VistaHeartBeatTimerManager
	 */
	protected VistaHeartBeatTimerManager getHeartBeatTimerManager() {
		return timerManager;
	}

	/**
	 * @return
	 * @va.exclude
	 */
	public String getAccessCode() {
		return accessCode;
	}

	/**
	 * @return
	 * @va.exclude
	 */
	public String getVerifyCode() {
		return verifyCode;
	}

	/**
	 * used for non-managed mode only
	 * 
	 * @param string
	 */
	public void setNonManagedAccessCode(String string) {
		accessCode = string;
		// additionally this must be a J2SE adapter
		setJ2SEProperties();
	}

	/**
	 * used for non-managed mode only
	 * 
	 * @param string
	 */
	public void setNonManagedVerifyCode(String string) {
		verifyCode = string;
		// additionally this must be a J2SE adapter
		setJ2SEProperties();
	}

	/**
	 * <ol>
	 * <li>Takes the raw IP address
	 * <li>Resolves it (especially needed if it is an unresolved name)
	 * <li>Sets the internal hostIpAddressResolved property to the resolved ip address
	 * </ol>
	 * 
	 * @param address Raw IP address to resolve (either machine name or textual representation of IP address)
	 */
	private void setHostIpAddressResolved(String address) throws UnknownHostException {
		if (hostIpAddressResolved != null) {
			return;
		}
		try {
			hostIpAddressResolved = InetAddress.getByName(address);
		} catch (UnknownHostException e) {
			throw e;
		}
	}

	/**
	 * @return
	 */
	public InetAddress getHostIpAddressResolved() {
		return hostIpAddressResolved;
	}

	/**
	 * 
	 * @return unresolved (original) host ip address setting
	 */
	public String getHostIpAddressUnresolved() {
		return this.hostIpAddress;
	}

	/**
	 * @return adapter environment
	 */
	public EMAdapterEnvironment getAdapterEnvironment() {
		return adapterEnvironment;
	}

	/**
	 * 
	 * @param adapterEnvironment
	 */
	public void setAdapterEnvironment(EMAdapterEnvironment adapterEnvironment) {
		this.adapterEnvironment = adapterEnvironment;
		this.timerManager.setAdapterEnvironment(this.adapterEnvironment);
	}

	/**
	 * @return
	 */
	protected long getManagedConnectionCount() {
		return managedConnectionCount;
	}

	/**
	 * @return
	 * @va.exclude
	 */
	public long getDistinguishedIdentifier() {
		return distinguishedIdentifier;
	}

	/**
	 * Setter used by container when processing custom ra.xml properties
	 * 
	 * @param connectorJndiName Jndi Name of the entry in the configuration file to use for this connector
	 */
	public void setConnectorJndiName(String connectorJndiName) {
		this.connectorJndiName = connectorJndiName;
		// if this property is being set, this must be a J2EE adapter
		this.setAdapterEnvironment(EMAdapterEnvironment.J2EE);
		this.parseConfigFailureMessage = null;
		try {
			ConfigurationReader.parseConfig(this);
			if (logger.isDebugEnabled()) {
				StringBuffer sb = new StringBuffer("connector Config JndiName: ");
				sb.append(connectorJndiName);
				sb.append("connector Config IP: ");
				sb.append(this.hostIpAddress);
				sb.append("connector Config Port: ");
				sb.append(this.hostPort);
				sb.append("connector Config Access code: <masked>\nconnector Config Verify code: <masked>");
				sb.append("connector Config Status: ");
				sb.append(this.connectorEnabled);
				logger.debug(sb.toString());
			}
		} catch (VistaLinkResourceException e) {
			this.parseConfigFailureMessage = e.getMessage();
			String loggerMsg = getLoggerFormattedStringWStackTrace(
					"Resource exception occurred in creatingManagedConnection()", e);
			logger.error(loggerMsg);
		}
	}

	/**
	 * overload setter method needed for JBoss
	 * 
	 * @param connectorJndiName
	 */
	public void setconnectorJndiName(String connectorJndiName) {
		setConnectorJndiName(connectorJndiName);
	}

	/**
	 * 
	 * @return Jndi name of the entry in the configuration file used for this connector
	 */
	public String getConnectorJndiName() {
		return this.connectorJndiName;
	}

	private void setJ2SEProperties() {
		if (!this.adapterEnvironment.equals(EMAdapterEnvironment.J2SE)) {
			this.setAdapterEnvironment(EMAdapterEnvironment.J2SE);
		}
		this.connectorEnabled = true;
	}

	// "friendly" scope
	void setAlwaysUseDefaultAsMin(boolean value) {
		alwaysUseDefaultAsMin = value;
	}

	/**
	 * 
	 * @return whether to use the default J2EE socket timeout as the minimum floor, even if timeout has been
	 *         programmatically set lower.
	 * 
	 * @va.exclude
	 */
	public boolean isAlwaysUseDefaultAsMin() {
		return alwaysUseDefaultAsMin;
	}

	void setDefaultJ2EETimeOut(long timeOut) {
		if (timeOut > J2EE_DEFAULT_SOCKET_TIMEOUT)
			defaultJ2EETimeOut = timeOut;
		else {
			defaultJ2EETimeOut = J2EE_DEFAULT_SOCKET_TIMEOUT;
			if (logger.isDebugEnabled())
				logger.debug(getLoggerFormattedString(new StringBuffer().append(
						"TimeOut configuration attribute value is too short [").append(timeOut).append(
						"ms]. The following default value is being used instead: ").append(J2EE_DEFAULT_SOCKET_TIMEOUT)
						.append("ms").toString()));
		}
	}

	/**
	 * @return current default J2EE socket timeout
	 * 
	 * @va.exclude
	 */
	public long getDefaultJ2EETimeOut() {
		return defaultJ2EETimeOut;
	}

	/**
	 * increment the connnection failure count
	 *  
	 */
	protected void incrementConnectionFailureCount() {
		synchronized (healthSyncObj) {
			this.connectionFailureCount++;
		}
	}

	/**
	 * health indicator
	 * 
	 * @return connection failure count
	 */
	public long getConnectionFailureCount() {
		synchronized (healthSyncObj) {
			return this.connectionFailureCount;
		}
	}

	/**
	 * increment the connection authentication failure count
	 *  
	 */
	protected void incrementConnectionAuthFailureCount() {
		synchronized (healthSyncObj) {
			this.connectionAuthFailureCount++;
		}
	}

	/**
	 * health indicator
	 * 
	 * @return the connection authentication failure count
	 */
	public long getConnectionAuthFailureCount() {
		synchronized (healthSyncObj) {
			return this.connectionAuthFailureCount;
		}
	}

	/**
	 * increment the identity failure count
	 *  
	 */
	protected void incrementIdentityFailureCount() {
		synchronized (healthSyncObj) {
			this.identityFailureCount++;
		}
	}

	/**
	 * health indicator
	 * 
	 * @return identity failure count
	 */
	public long getIdentityFailureCount() {
		synchronized (healthSyncObj) {
			return this.identityFailureCount;
		}
	}

	/**
	 * increment the division mismatch count
	 *  
	 */
	protected void incrementdivisionMismatchCount() {
		synchronized (healthSyncObj) {
			this.divisionMismatchCount++;
		}
	}

	/**
	 * health indicator
	 * 
	 * @return division mismatch count
	 */
	public long getDivisionMismatchCount() {
		synchronized (healthSyncObj) {
			return this.divisionMismatchCount;
		}
	}

	/**
	 * increment the production mismatch count
	 *  
	 */
	protected void incrementProductionMismatchCount() {
		synchronized (healthSyncObj) {
			this.productionMismatchCount++;
		}
	}

	/**
	 * @param connectorEnabled The connectorEnabled to set.
	 */
	protected void setConnectorEnabled(boolean connectorEnabled) {
		this.connectorEnabled = connectorEnabled;
	}

	/**
	 * health indicator
	 * 
	 * @return production mismatch count
	 */
	public long getProductionMismatchCount() {
		synchronized (healthSyncObj) {
			return this.productionMismatchCount;
		}
	}

	/**
	 * Add a new time for creating another connection handle
	 * 
	 * @param elapsedMillis time to create connection handle, in milliseconds
	 */
	protected void addCreateConnectionHandleTime(long elapsedMillis) {
		synchronized (createConnectionHandleAvgSyncObj) {
			if (logger.isDebugEnabled()) {
				logger.debug("elapsedMillis to add: " + Long.toString(elapsedMillis));
			}
			//TODO set some kind of flag for console if reach MAX_VALUE limit? reset?
			if (elapsedMillis < (Long.MAX_VALUE - this.createConnectionHandleTotalMillis)) {
				this.createConnectionHandleCount++;
				this.createConnectionHandleTotalMillis = this.createConnectionHandleTotalMillis + elapsedMillis;
			} else {
				logger.error("CreateConnectionHandle time total exceeded LONG MAX_VALUE");
			}
		}
	}

	/**
	 * health indicator
	 * 
	 * @return average time to create connection handle, in milliseconds
	 */
	public double getCreateConnectionHandleAvgMillis() {
		synchronized (createConnectionHandleAvgSyncObj) {
			if (this.createConnectionHandleCount > 0) {
				return ((new Long(createConnectionHandleTotalMillis)).doubleValue())
						/ ((new Long(this.createConnectionHandleCount)).doubleValue());
			} else {
				return 0.0;
			}
		}
	}

	/**
	 * update the match managed connection average time
	 * 
	 * @param elapsedMillis new match managed connection time, in milliseconds
	 */
	private void addMatchManagedConnectionTime(long elapsedMillis) {
		synchronized (this.matchManagedConnectionAvgSyncObj) {
			//TODO set some kind of flag for console if reach MAX_VALUE limit? reset?
			if (elapsedMillis < (Long.MAX_VALUE - this.matchManagedConnectionTotalMillis)) {
				this.matchManagedConnectionCount++;
				this.matchManagedConnectionTotalMillis = this.matchManagedConnectionTotalMillis + elapsedMillis;
			} else {
				logger.error("MatchManagedConnection time total exceeded LONG MAX_VALUE");
			}
		}
	}

	/**
	 * health indicator
	 * 
	 * @return match managed connection average time in milliseconds
	 */
	public double getMatchManagedConnectionAvgMillis() {
		synchronized (this.matchManagedConnectionAvgSyncObj) {
			if (this.matchManagedConnectionCount > 0) {
				return ((new Long(this.matchManagedConnectionTotalMillis)).doubleValue() / (new Long(
						this.matchManagedConnectionCount)).doubleValue());
			} else {
				return 0.0;
			}
		}
	}

	/**
	 * @return Returns the primaryStation.
	 */
	protected String getPrimaryStation() {
		return primaryStation;
	}

	/**
	 * @param accessCode The accessCode to set.
	 */
	protected void setAccessCode(String accessCode) {
		this.accessCode = accessCode;
	}

	/**
	 * @param hostIpAddress The hostIpAddress to set.
	 */
	protected void setHostIpAddress(String hostIpAddress) {
		this.hostIpAddress = hostIpAddress;
	}

	/**
	 * @param hostPort The hostPort to set.
	 */
	protected void setHostPort(int hostPort) {
		this.hostPort = hostPort;
	}

	/**
	 * @param primaryStation The primaryStation to set.
	 */
	protected void setPrimaryStation(String primaryStation) {
		this.primaryStation = primaryStation;
	}

	/**
	 * @param primaryStationSuffix The primaryStationSuffix to set.
	 */
	protected void setPrimaryStationSuffix(String primaryStationSuffix) {
		this.primaryStationSuffix = primaryStationSuffix;
	}

	/**
	 * @param verifyCode The verifyCode to set.
	 */
	protected void setVerifyCode(String verifyCode) {
		this.verifyCode = verifyCode;
	}

	/**
	 * Finalizer to unmap institution mappings at some point after connector is undeployed by container.
	 */
	protected void finalize() throws Throwable {
		if (this.adapterEnvironment == EMAdapterEnvironment.J2EE) {
			try {
				if (logger.isDebugEnabled()) {
					logger.debug("considering removing mappings for JNDI Name '" + connectorJndiName
							+ "', distinguished identifier '" + distinguishedIdentifier + ".");
				}
				InstitutionMappingImpl instMap = InstitutionMappingImpl.getInstance();
				instMap.removeMappingsForJndiName(connectorJndiName, new String[] { primaryStation },
						distinguishedIdentifier);
			} finally {
				super.finalize();
			}
		}
	}

}