package gov.va.med.vistalink.adapter.spi;

import gov.va.med.vistalink.adapter.cci.VistaLinkConnectionFactory;
import gov.va.med.vistalink.institution.InstitutionMappingImpl;
import gov.va.med.vistalink.institution.InstitutionMappingVO;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.QueryExp;
import javax.management.ReflectionException;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.Logger;

import weblogic.management.MBeanHome;
import weblogic.management.configuration.ApplicationMBean;
import weblogic.management.configuration.ConnectorComponentMBean;
import weblogic.management.runtime.ApplicationRuntimeMBean;
import weblogic.management.runtime.ConnectorComponentRuntimeMBean;
import weblogic.management.runtime.ServerRuntimeMBean;

/**
 * One instance of this MBean is created per JVM, by a static initializer in the VistaLinkManagedConnectionFactory
 * class. The MBean is used by the VistaLink console to provide information on VistaLink deployed connectors, using the
 * WebLogic MBean hierarchy to obtain the list of WebLogic-deployed connectors.
 * 
 * @author Application Modernization - Foundations Team
 * @version 1.5.0.026
 */
public class VistaLinkRuntimeInfo implements VistaLinkRuntimeInfoMBean {

	private static final Logger logger = Logger.getLogger(VistaLinkRuntimeInfo.class);

	private ObjectName objectName;
	private MBeanHome localHome;
	private String domainName;
	private String localServerName;

	/**
	 * Constructor.
	 * @throws Exception
	 * @va.exclude
	 */
	public VistaLinkRuntimeInfo() throws NamingException {

		logger.debug("in constructor");
		// initialize values
		initialize();

	}

	/**
	 * Initialize the instance.
	 * @throws Exception
	 */
	private void initialize() throws NamingException {
		try {
			InitialContext ctx = new InitialContext();
			/*
			 * Note: we get this rather than MBeanHome because we're interacting with non-WebLogic Server MBeans. And,
			 * we get local server home because we want MBean on same server that connector is deployed on.
			 */
			this.localHome = (MBeanHome) ctx.lookup(MBeanHome.LOCAL_JNDI_NAME);
			this.domainName = localHome.getDomainName();
			this.localServerName = getServerName(localHome);
		} catch (NamingException e) {
			logger.error("Initialization failed: ", e);
			throw e;
		}
	}

	/**
	 * @return the local server name
	 */
	private static String getServerName(MBeanHome localHome) {
		String returnVal = null;
		Set serverMBeans = localHome.getMBeansByType("ServerRuntime");
		Iterator iter = serverMBeans.iterator();
		while (iter.hasNext()) {
			// using local home, should be only one
			returnVal = ((ServerRuntimeMBean) iter.next()).getName();
			break;
		}
		return returnVal;
	}

	//MBEAN interface method
	public Map retrieveConnectorList() {

		//todo throw exception here?
		TreeMap returnVal = new TreeMap();

		MBeanServer localMBeanServer = localHome.getMBeanServer();
		Set connectorMBeanSet = localHome.getMBeansByType("ConnectorComponentRuntime");
		if (logger.isDebugEnabled()) {
			logger.debug("found " + connectorMBeanSet.size() + " ConnectorComponentRuntime mbeans");
		}
		// loop through ConnectorComponentRuntime mbeans
		for (Iterator connectorIter = connectorMBeanSet.iterator(); connectorIter.hasNext();) {

			ConnectorInfoVO cisVO = new ConnectorInfoVO();
			ConnectorComponentRuntimeMBean runtimeMBean = (ConnectorComponentRuntimeMBean) connectorIter.next();
			if (logger.isDebugEnabled()) {
				logger.debug("retrieving info for ConnectorComponentRuntimeMBean: '"
						+ runtimeMBean.getObjectName().getCanonicalName() + "'.");
			}
			getConnectorInfo(cisVO, runtimeMBean, false);
			// only add to set if it's a VISTA EIS type, or not-yet-determined type
			if ("VistA".equals(cisVO.getWlsEisType()) || (cisVO.getWlsEisType() == null)) {
				returnVal.put(runtimeMBean.getObjectName().getCanonicalName(), cisVO);
			} else {
				// we're skipping because it's explicitly identified as not a VistALink connector
				if (logger.isDebugEnabled()) {
					logger.debug("skipping connector EisType: '" + cisVO.getWlsEisType() + "'.");
				}
			}
		}

		if (logger.isDebugEnabled()) {
			logger.debug("Returning connector list with " + returnVal.size() + " entries.");
		}
		return returnVal;
	}

	/**
	 * Helper method to retrieve VistaLink "connectorJndiName" custom ra.xml attribute from a Connector configuration
	 * mbean
	 * @param configMBean connector configuration mbean to get connectorJndiName attribute from
	 * @return
	 */
	private String getVistaLinkConfigJndiName(ConnectorComponentMBean configMBean) {

		String returnVal = null;

		try {
			// look specifically for VistALink connectorJndiName custom ra.xml attribute
			logger.debug("About to get ConfigProperties Hashtable");
			Hashtable configProperties = (Hashtable) configMBean.getAttribute("ConfigProperties");
			logger.debug("about to get " + VistaLinkManagedConnectionFactory.CONNECTOR_PROPERTY_NAME
					+ " from ConfigProperties");
			if (configProperties == null) {
				logger.debug("... but ConfigProperties is null.");
			} else if (!configProperties.containsKey(VistaLinkManagedConnectionFactory.CONNECTOR_PROPERTY_NAME)) {
				logger.debug("... but ConfigProperties doesn't contain '"
						+ VistaLinkManagedConnectionFactory.CONNECTOR_PROPERTY_NAME + "' key");
			} else {
				logger.debug("configProperties contained the key '"
						+ VistaLinkManagedConnectionFactory.CONNECTOR_PROPERTY_NAME + "'");
				Hashtable ht1 = (Hashtable) configProperties
						.get(VistaLinkManagedConnectionFactory.CONNECTOR_PROPERTY_NAME);
				if (ht1 == null) {
					logger.debug("could not retrieve " + VistaLinkManagedConnectionFactory.CONNECTOR_PROPERTY_NAME
							+ " inner Hashtable, it is null");
				} else {
					returnVal = (String) ht1.get("Value");
					if (logger.isDebugEnabled()) {
						logger.debug(" found " + VistaLinkManagedConnectionFactory.CONNECTOR_PROPERTY_NAME + " = "
								+ returnVal);
					}
				}
			}
		} catch (AttributeNotFoundException e) {
			logger.error(e);
		} catch (MBeanException e) {
			logger.error(e);
		} catch (ReflectionException e) {
			logger.error(e);
		}

		return returnVal;
	}

	//MBEAN interface method
	public ConnectorInfoVO retrieveConnectorDetailInfo(String canonicalRuntimeMBeanObjectName) {
		//todo initialize to null instead?
		//todo throw exception here?
		ConnectorInfoVO cisVO = new ConnectorInfoVO();
		try {
			ConnectorComponentRuntimeMBean runtimeMBean = (ConnectorComponentRuntimeMBean) localHome
					.getMBean(new ObjectName(canonicalRuntimeMBeanObjectName));
			getConnectorInfo(cisVO, runtimeMBean, true);
		} catch (InstanceNotFoundException e) {
			logger.warn("Could not find MBean for '" + canonicalRuntimeMBeanObjectName
					+ "'; perhaps the connector was undeployed.");
		} catch (MalformedObjectNameException e) {
			logger.error(e);
		} catch (NullPointerException e) {
			logger.error(e);
		}
		return cisVO;
	}

	/**
	 * Populates a ConnectorInfoVO for either summary or detail mode
	 * @param cisVO ConnectorInfoVO to populate
	 * @param runtimeMBean ConnectorComponentRuntimeMBean for the connector in question
	 * @param getDetailInfo whether to get detailed info (more than summary fields) or not
	 */
	private void getConnectorInfo(ConnectorInfoVO cisVO, ConnectorComponentRuntimeMBean runtimeMBean,
			boolean getDetailInfo) {

		cisVO.setWlsDeploymentState(runtimeMBean.getDeploymentState());
		cisVO.setWlsDeploymentStateString(getWlsDeploymentStateName(runtimeMBean.getDeploymentState()));

		// get WLS information from ConnectorComponent config mbean for the connector
		ConnectorComponentMBean matchingConfigMBean = getConnectorConfigMBean(runtimeMBean);
		if (matchingConfigMBean == null) {
			logger.debug("Did not find a matching ConnectorComponent config mbean.");
		} else {
			logger.debug("got matching config mbean for runtime.");
			try {
				// get Name (should be defined)
				// String appName = (String) matchingConfigMBean.getAttribute("Name");
				String appName = matchingConfigMBean.getObjectName().getKeyProperty("ApplicationConfig");

				if (appName != null) {
					cisVO.setWlsApplicationName(appName);
				}

				// get EisType (may be null)
				String eisType = (String) matchingConfigMBean.getAttribute("EisType");
				if (eisType != null) {
					cisVO.setWlsEisType(eisType);

					// don't do the rest if EISType not VistA
					if (eisType.equals("VistA")) {

						cisVO.setWlsEisType("VistA");
						// we know it's VistALink now
						// get JndiName (may be null)
						String deployedJndiName = (String) matchingConfigMBean.getAttribute("JndiName");

						if (deployedJndiName != null) {

							// if null, more to do -- deployment is not far enough along yet. Won't have access to
							// the "config" VistaLink JNDI name to a) see if mismatch, b) definitively pick
							// correct config entry, instead of assuming there's no mismatch and picking the
							// deployed JNDI name to retrieve entry with

							cisVO.setWlsJndiActual(deployedJndiName);
							cisVO.setCfgJndiName(getVistaLinkConfigJndiName(matchingConfigMBean));

							// get connection factory via live JNDI lookup, so we get the real CF
							// associated with this connector. Use to retrieve MCF-stored values.
							logger.debug("about to get CF");
							VistaLinkConnectionFactory cf = getCF(deployedJndiName, runtimeMBean);
							if (logger.isDebugEnabled()) {
								logger.debug("got CF: " + cf);
							}
							if (cf != null) {
								// would be null if deployment state is not ACTIVATED
								cf.getMCFInfo(cisVO, getDetailInfo);
							}

							if (getDetailInfo) {
								cisVO.setWlsConnectionFactoryName((String) matchingConfigMBean
										.getAttribute("ConnectionFactoryName"));
								cisVO.setWlsResourceAdapterLinkRef((String) matchingConfigMBean
										.getAttribute("RALinkRef"));
								cisVO.setCfgStationNumbers(getStationNumbers(deployedJndiName));
							}
							logger.debug("done with MCF");
						}
					}
				}

			} catch (AttributeNotFoundException e) {
				logger.error(e);
			} catch (MBeanException e) {
				logger.error(e);
			} catch (ReflectionException e) {
				logger.error(e);
			} catch (Exception e) {
				logger.error(e);
			}
		}
	}

	/**
	 * Helper method to get the list of station#s for a particular connector, based on the current mappings in place.
	 * @param deployedJndiName
	 * @return
	 */
	private String[] getStationNumbers(String deployedJndiName) {

		String[] stationNumbers = null;
		logger.debug("looking for live mappings for jndi name # '" + deployedJndiName + "'.");

		// get Institution Mappings
		InstitutionMappingImpl me = InstitutionMappingImpl.getInstance();
		Map institutionMap = me.getMappingClone();
		// iterate through all stations, find any matching jndi names. We use the deployed JNDI name (rather
		// than the config JNDI name) so that we reflect how the mapping is actually working for applications
		// hitting this connector.
		TreeSet matchingStations = new TreeSet();
		Set keys = institutionMap.keySet();
		for (Iterator iter = keys.iterator(); iter.hasNext();) {
			String stationNumber = (String) iter.next();
			InstitutionMappingVO imVO = (InstitutionMappingVO) institutionMap.get(stationNumber);
			String mappingJndiName = imVO.getJndiName();
			if (mappingJndiName.equals(deployedJndiName)) {
				matchingStations.add(stationNumber);
				logger.debug("found matching station # '" + stationNumber + "'.");
			}
		}
		// store any matched institutions
		if (matchingStations.size() > 0) {
			stationNumbers = new String[matchingStations.size()];
			int i = 0;
			for (Iterator iter = matchingStations.iterator(); iter.hasNext();) {
				stationNumbers[i] = (String) iter.next();
				i++;
			}
		}
		return stationNumbers;
	}

	/**
	 * Helper method to retrieve connection factory based on a jndi name.
	 * @param jndiName Name to use to retrieve connection factory.
	 * @param wlsDeploymentState WLS deployment state, used to determine whether to log error
	 * @return Connection factory, based on live jndi lookup.
	 */
	private VistaLinkConnectionFactory getCF(String jndiName, ConnectorComponentRuntimeMBean runtimeMBean) {

		VistaLinkConnectionFactory cf = null;
		try {
			InitialContext ctx = new InitialContext();
			cf = (VistaLinkConnectionFactory) ctx.lookup(jndiName);
		} catch (NamingException e) {
			int deploymentState = runtimeMBean.getDeploymentState();
			if (deploymentState == ConnectorComponentRuntimeMBean.ACTIVATED) {
				logger.error(e);
			} else {
				logger.debug("Encountered NamingException trying to retrieve connection factory '" + jndiName
						+ "' from JNDI, but deployment state is " + getWlsDeploymentStateName(deploymentState));
			}
		}
		return cf;
	}

	/**
	 * Helper method, to retrieve a connector configuration MBean, given a ConnectorComponentRuntimeMBean.
	 * @param runtimeMBean
	 * @return ConnectorConfigMBean that is the config MBean for the runtime MBean passed in
	 */
	private ConnectorComponentMBean getConnectorConfigMBean(ConnectorComponentRuntimeMBean runtimeMBean) {

		ConnectorComponentMBean matchingConfigBean = null;
		try {

			// get appRuntimeMBean
			ApplicationRuntimeMBean appRuntimeMBean = (ApplicationRuntimeMBean) runtimeMBean.getParent();
			if (appRuntimeMBean == null) {
				logger.error("Could not get matching ApplicationRuntimeMBean");
			} else {

				// get appConfigMBean MBean
				String appName = (String) appRuntimeMBean.getAttribute("ApplicationName");
				if (appName == null) {
					logger.error("could not get ApplicationName attribute.");
				} else {
					String domainName = localHome.getDomainName();
					StringBuffer nameBuffer = new StringBuffer(domainName);
					nameBuffer.append(":Location=").append(localServerName).append(",Name=").append(appName).append(
							",Type=ApplicationConfig");
					if (logger.isDebugEnabled()) {
						logger.debug("Looking for ObjectName: '" + nameBuffer.toString() + "'.");
					}
					ApplicationMBean appConfigMBean = (ApplicationMBean) localHome.getMBean(new ObjectName(nameBuffer
							.toString()));
					if (appConfigMBean == null) {
						logger.debug("could not retrieve matching ApplicationConfigMBean.");
					} else {
						logger.debug("found matching ApplicationConfigMBean.");

						// get ConnectorComponentConfig MBean
						QueryExp query = null;
						MBeanServer localMBeanServer = this.localHome.getMBeanServer();
						nameBuffer = new StringBuffer(domainName);
						nameBuffer.append(":ApplicationConfig=").append(appName);
						nameBuffer.append(",Location=").append(localServerName);
						nameBuffer.append(",Type=ConnectorComponentConfig,*");
						if (logger.isDebugEnabled()) {
							logger.debug("Looking for ObjectName: '" + nameBuffer.toString() + "'.");
						}
						Set connectorConfigBeanObjectNames = localMBeanServer.queryMBeans(new ObjectName(nameBuffer
								.toString()), query);

						// scan set of matching ConnectorComponentConfig MBeans (should only be one, as far as can tell)
						for (Iterator iter = connectorConfigBeanObjectNames.iterator(); iter.hasNext();) {
							ObjectInstance objInstance = (ObjectInstance) iter.next();
							if (!("weblogic.management.configuration.ConnectorComponentMBean".equals(objInstance
									.getClassName()))) {
								logger
										.error("Expected to find object of class ConnectorComponentMBean, instead found: "
												+ objInstance.getClassName());
							} else {

								// check if ConnectorComponentConfig parent = previously found AppConfigMBean. If so --
								// it's a match.
								ObjectName oName = objInstance.getObjectName();
								ConnectorComponentMBean candidateConfigMBean = (ConnectorComponentMBean) localHome
										.getMBean(oName);
								if (candidateConfigMBean == null) {
									logger.error("Could not retrieve ConectorComponentMBean for ObjectName: "
											+ oName.getCanonicalName());
								} else {

									if (candidateConfigMBean.getParent() == appConfigMBean) {
										if (logger.isDebugEnabled()) {
											logger
													.debug("found parent match! ConnectorComponentConfigMBean object name: "
															+ candidateConfigMBean.getObjectName());
										}
										matchingConfigBean = candidateConfigMBean;
										break;
									} else {
										if (logger.isDebugEnabled()) {
											logger.debug("Not a match: ConnectorComponentConfigMBean object name: "
													+ candidateConfigMBean.getObjectName());
										}
									}
								}
							}
						}
					}
				}
			}
		} catch (MalformedObjectNameException e) {
			logger.error(e);
		} catch (AttributeNotFoundException e) {
			logger.error(e);
		} catch (NullPointerException e) {
			logger.error(e);
		} catch (MBeanException e) {
			logger.error(e);
		} catch (ReflectionException e) {
			logger.error(e);
		} catch (Exception e) {
			logger.error(e);
		}

		return matchingConfigBean;
	}

	/**
	 * toString.
	 */
	public String toString() {
		return objectName.toString();
	}

	/**
	 * Utility method for users outside this class, to create a canonical object name to retrieve an instance of this
	 * MBean on a given server. Also used internally to create the ObjectName the runtime object is registered under.
	 * @param domainName domain to use in naming of MBean
	 * @param serverName server name to use in naming of MBean
	 * @return canonical object name string for use in constructing ObjectName for MBean
	 * @va.exclude
	 */
	public static String getServerSpecificCanonicalObjectName(String domainName, String serverName) {

		StringBuffer sb = new StringBuffer(domainName).append(":Name=").append(serverName).append("-").append(
				"VistaLinkRuntimeInfo,Type=VistaLinkRuntimeInfo,Location=").append(serverName);
		return sb.toString();
	}

	/**
	 * Creates/registers the MBean instance.
	 * @va.exclude
	 */
	static public VistaLinkRuntimeInfo register() {

		ObjectName mbeanObjectName = null;
		VistaLinkRuntimeInfo monitor = null;

		// catch all Exceptions, or else will cause static initializer errors in VistaLinkManagedConnectionFactory
		try {
			InitialContext ctx = new InitialContext();
			monitor = new VistaLinkRuntimeInfo();
			MBeanServer mbeanServer = monitor.localHome.getMBeanServer();
			String canonicalObjectName = getServerSpecificCanonicalObjectName(monitor.domainName,
					monitor.localServerName);
			mbeanObjectName = new ObjectName(canonicalObjectName);
			// register mcf mbean
			ObjectInstance mbeanInstance = mbeanServer.registerMBean((Object) monitor, mbeanObjectName);
			monitor.setObjectName(mbeanObjectName);
			if (logger.isDebugEnabled()) {
				logger.debug("MBean '" + mbeanObjectName + "' registered.");
			}

		} catch (Exception e) {
			logger.error("MBean initialization or registration failed for '" + mbeanObjectName + "'.", e);
		}

		return monitor;
	}

	//MBEAN interface method
	public ObjectName getObjectName() {
		return objectName;
	}

	/**
	 * @param objectName The ObjectName under which the MBean was registered.
	 */
	private void setObjectName(ObjectName objectName) {
		this.objectName = objectName;
	}

	/**
	 * Helper method to transform deploymentState int to human-friendly string
	 * @param deploymentState WLS deployment state for connector
	 * @return human-friendly string
	 */
	private String getWlsDeploymentStateName(int deploymentState) {
		String returnVal = null;
		if (deploymentState == ConnectorComponentRuntimeMBean.NEW) {
			returnVal = "NEW";
		} else if (deploymentState == ConnectorComponentRuntimeMBean.UNPREPARED) {
			returnVal = "UNPREPARED";
		} else if (deploymentState == ConnectorComponentRuntimeMBean.PREPARED) {
			returnVal = "PREPARED";
		} else if (deploymentState == ConnectorComponentRuntimeMBean.ACTIVATED) {
			returnVal = "ACTIVATED";
		}
		return returnVal;
	}
}