package gov.va.med.vistalink.adapter.spi;

import java.io.Serializable;
import java.util.HashMap;

import org.apache.log4j.Logger;

/**
 * This class represents a serialized repsonse to the VistaLinkSystemInfoRequest. The VistaLinkSystemInfoResponse class
 * represents the parsed version of the response.
 * <p>
 * The read only properties represent VistaLink system information for a M/VistA server.
 * <p>
 * This value object is needed becuase it must be serializable to go across JVMs for the admin console add-in.
 * 
 * @author Application Modernization - Foundations Team
 * @version 1.5.0.026
 */
public class VistaLinkSystemInfoVO implements Serializable {

	/**
	 * The logger used for this class
	 */
	private static final Logger logger = Logger.getLogger(VistaLinkSystemInfoVO.class);

	/**
	 * The VistaLink version on the M/VistA server
	 */
	private String vistalinkVersion;

	/**
	 * The Vista Kernel setting for whether the account is production or test
	 */
	private String vistaProduction;

	/**
	 * The default institution for the remote system from the Kernel System Parameters file.
	 */
	private String defaultInstitution;

	/**
	 * The VistaLink build number on the M/VistA server
	 */
	private String vistalinkBuild;

	/**
	 * The application server connection timeount (in milliseconds) set on the M/VistA server
	 */
	private long appSeverTimeout;

	/**
	 * The application server reauthenticated session timeout (in milliseconds) set on the M/VistA server
	 */
	private long reAuthSessionTimeout;

	/**
	 * The uci name of M/VistA server
	 */
	private String uci;

	/**
	 * The volume name of M/VistA server
	 */
	private String vol;

	/**
	 * The box-volume name of M/VistA server
	 */
	private String boxVolume;

	/**
	 * The M version name of M/VistA server
	 */
	private String mVersion;

	/**
	 * The operating system on which the M/VistA server runs under
	 */
	private String operatingSystem;

	/**
	 * The domain name of M/VistA server
	 */
	private String domainName;

	/**
	 * Location for object user to stash an error message when returning this object
	 */
	private String errorMessage;

	/**
	 * c/p username
	 */
	private String connectorProxyUsername;

	private String introText;

	/**
	 * no-arg constructor
	 */
	public VistaLinkSystemInfoVO() {
		super();
	}

	/**
	 *  
	 */
	public VistaLinkSystemInfoVO(HashMap hm) {
		super();

		// populate properties
		this.vistalinkVersion = (String) hm.get("vistalinkVersion");
		this.vistalinkBuild = (String) hm.get("vistalinkBuild");
		this.appSeverTimeout = Long.parseLong((String) hm.get("appServerTimeout"));
		this.reAuthSessionTimeout = Long.parseLong((String) hm.get("reAuthSessionTimeout"));
		this.uci = (String) hm.get("uci");
		this.vol = (String) hm.get("vol");
		this.boxVolume = (String) hm.get("boxVolume");
		this.mVersion = (String) hm.get("mVersion");
		this.operatingSystem = (String) hm.get("operatingSystem");
		this.domainName = (String) hm.get("domainName");
		this.vistaProduction = (String) hm.get("vistaProduction");
		this.defaultInstitution = (String) hm.get("defaultInstitution");
		this.introText = (String) hm.get("introductoryText");
		this.connectorProxyUsername = (String) hm.get("cpName");

		// log message for debugging
		if (logger.isDebugEnabled()) {
			logger.debug((new StringBuffer()).append("system info vo constructed"));
		}
	}

	/**
	 * @return Returns the vistalinkVersion.
	 */
	public String getVistalinkVersion() {
		return vistalinkVersion;
	}

	/**
	 * @return Returns the vistalinkBuild.
	 */
	public String getVistalinkBuild() {
		return vistalinkBuild;
	}

	/**
	 * @return Returns the application server connection listener time out
	 */
	public long getAppServerTimeout() {
		return appSeverTimeout;
	}

	/**
	 * @return Returns whether the remote M system is Production (true | false | unknown)
	 */
	public String getVistaProduction() {
		return this.vistaProduction;
	}

	/**
	 * @return Returns the default institution (station#) for the M system.
	 */
	public String getDefaultInstitution() {
		return this.defaultInstitution;
	}

	/**
	 * @return Returns the reauthenticated session time out
	 */
	public long getReAuthSessionTimeout() {
		return reAuthSessionTimeout;
	}

	/**
	 * @return Returns the uci name
	 */
	public String getUci() {
		return uci;
	}

	/**
	 * @return Returns the vol name
	 */
	public String getVol() {
		return vol;
	}

	/**
	 * @return Returns the box-volume name
	 */
	public String getBoxVolume() {
		return boxVolume;
	}

	/**
	 * @return Returns the M version name
	 */
	public String getMVersion() {
		return mVersion;
	}

	/**
	 * Returns the operating system name
	 * 
	 * @return long
	 */
	public String getOperatingSystem() {
		return operatingSystem;
	}

	/**
	 * @return Returns the domainName.
	 */
	public String getDomainName() {
		return domainName;
	}

	/**
	 * @return Returns an errorMessage store in the VO by the user of the VO.
	 */
	public String getErrorMessage() {
		return errorMessage;
	}

	/**
	 * @param errorMessage
	 *            The errorMessage to set.
	 */
	public void setErrorMessage(String errorMessage) {
		this.errorMessage = errorMessage;
	}

	/**
	 * @return Returns the introText.
	 */
	public String getIntroText() {
		return introText;
	}

	/**
	 * @param introText
	 *            The introText to set.
	 */
	public void setIntroText(String introText) {
		this.introText = introText;
	}

	/**
	 * @return connector proxy username
	 */
	public String getCpName() {
		return this.connectorProxyUsername;
	}

	/**
	 * @param cpName
	 *            connector proxy username
	 */
	public void setCpName(String cpName) {
		this.connectorProxyUsername = cpName;
	}

	public String toString() {
		return "VistaLink version: '" + vistalinkVersion + "' build: '" + vistalinkBuild
				+ "' Application Sever Timeout: '" + appSeverTimeout + "' UCI: '" + uci + "' VOL: '" + vol
				+ "' BOX-VOLUME: '" + boxVolume + "' M Version: ' " + mVersion + "' Operating System: '"
				+ operatingSystem + "' Domain: " + domainName + "' Production: '" + vistaProduction
				+ "' Default Institution Station#: '" + defaultInstitution + "'";
	}
}