package gov.va.med.vistalink.institution;

import java.util.Set;

import org.apache.log4j.Logger;

/**
 * The in-memory mapping is initialized during connector startup.
 * 
 * @author Application Modernization - Foundations Team
 * @version 1.5.0.026
 */
public class InstitutionMappingDelegate {

	private static Logger logger = Logger.getLogger(InstitutionMappingDelegate.class);

	/**
	 * Returns the JNDI connector name (if found) for a division matching the station number passed in.
	 * <p>
	 * An example of retrieving the JNDI name for a connector is:
	 * <p>
	 * <code>
	 * String stationNumber = 500;
	 * String jndiConnectorName = null;<br>
	 * try {<br>
	 * 	jndiConnectorName = InstitutionMappingDelegate.getJndiConnectorNameForInstitution(stationNumber);<br>
	 * } catch (InstitutionMappingNotFoundException e) {<br>
	 * 	// take some action<br>
	 * } catch (InstitutionMapNotInitializedException e) {<br>
	 * 	// take some action<br>
	 * }
	 * </code>
	 * 
	 * @param stationNumber
	 *            institution station number for Vista Institution.
	 * @return Jndi Connector name, if a connector is found that is mapped to input station number. Use for a JNDI
	 *         lookup of actual VistaLink connector.
	 * @throws InstitutionMapNotInitializedException
	 *             thrown if the institution mapping has not been initialized
	 * @throws InstitutionMappingNotFoundException
	 *             thrown if no connector JNDI name has been mapped to input station number
	 */
	public static String getJndiConnectorNameForInstitution(String stationNumber)
			throws InstitutionMapNotInitializedException, InstitutionMappingNotFoundException {
		InstitutionMapping im = InstitutionMappingFactory.getInstitutionMapping();
		if (im == null) {
			throw new InstitutionMapNotInitializedException();
		} else {

			// strip down to numeric portion only
			int nextDigitPosToTest = 0;
			while ((nextDigitPosToTest < stationNumber.length())
					&& (Character.isDigit(stationNumber.charAt(nextDigitPosToTest)))) {
				nextDigitPosToTest++;
			}
			//nextDigitPosToTest-1 has last digit position for first numeric
			// portion of string
			if (nextDigitPosToTest > 0) {
				// nursing home
				if ((nextDigitPosToTest == 4) && (stationNumber.charAt(3) == '9')) {
					return im.getJndiConnectorNameForInstitution(stationNumber.substring(0, 3));
				}
				// AAC suffixes if present
				if ((stationNumber.charAt(0) == '2') && (stationNumber.length() > nextDigitPosToTest)) {
					// look up with suffix first; if fails, fall through to
					// normal case
					String returnVal = null;
					try {
						returnVal = im.getJndiConnectorNameForInstitution(stationNumber);
						return returnVal;
					} catch (Exception e) {
						if (logger.isDebugEnabled()) {
							logger.debug("Tried to look up AAC-type station# with suffix attached, failed: '"
									+ stationNumber + "'.");
						}
					}
				}
				// normal case -- truncate suffix, look up numeric portion only.
				return im.getJndiConnectorNameForInstitution(stationNumber.substring(0, nextDigitPosToTest));
			} else {
				throw new InstitutionMappingNotFoundException("Invalid station number format: '" + stationNumber + "'.");
			}
		}
	}

	/**
	 * Returns the set of station# strings for which adapters have been deployed on the current JVM, and for which
	 * institution mappings currently exist. The source of the station#s in the mapping is the primaryStation
	 * configuration attribute, for successfully deployed adapters.
	 * @return Set containing station# strings.
	 * @throws InstitutionMapNotInitializedException
	 */
	public static Set getVistaLinkMappedStationNumberSet() throws InstitutionMapNotInitializedException {

		Set returnVal = null;
		InstitutionMapping im = InstitutionMappingFactory.getInstitutionMapping();
		if (im == null) {
			throw new InstitutionMapNotInitializedException();
		} else {
			returnVal = im.getVistaLinkMappedStationNumberSet();
		}
		return returnVal;
	}
}