package gov.va.med.vistalink.institution;

import gov.va.med.vistalink.adapter.cci.VistaLinkResourceException;

import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import org.apache.log4j.Logger;
import org.apache.log4j.Priority;

/**
 * The implementation of the singleton object used to hold the in-memory mapping of institutions to VistaLink connector
 * JNDI names.
 * <p>
 * The singleton instance of this class is initialized during connector startup.
 * <p>
 * This class is thread-safe.
 * 
 * @author Application Modernization - Foundations Team
 * @version 1.5.0.026
 * @va.exclude
 */
public class InstitutionMappingImpl implements InstitutionMapping {

	private static final Logger logger = Logger.getLogger(InstitutionMappingImpl.class);

	/*
	 * internal object to keep track of singleton instance
	 */
	private static InstitutionMappingImpl me = null;

	/*
	 * internal list of institutions, read from config file. Key = station#, value = jndi name.
	 */
	private TreeMap institutionMap = new TreeMap();

	/*
	 * synchronize creation of singleton
	 */
	private static Object syncObjSingleton = new Object();

	/*
	 * synchronize concurrent access to mappings
	 */
	private Object syncObjMapping = new Object();

	/**
	 * Returns a reference to the instance of the singleton object for this class. If the object has not been
	 * instantiated, null is returned, as this object cannot be constructed/initialized without additional information,
	 * namely the list of division mappings that the object depends on.
	 * 
	 * @return a reference to the instance of the singleton object, or null if it hasn't been constructed.
	 * @throws IllegalStateException
	 *             thrown if object does not exist yet
	 * @va.exclude
	 */
	public static InstitutionMappingImpl getInstance() throws IllegalStateException {
		if (logger.isEnabledFor(Priority.DEBUG)) {
			logger.debug("InstitutionMappingImpl :: In getInstance().");
		}

		synchronized (syncObjSingleton) {
			if (me == null) {
				String errStr = "InstitutionMappingImpl :: Instance does not exist.";
				if (logger.isEnabledFor(Priority.ERROR)) {
					logger.error(errStr);
				}
				throw new IllegalStateException(errStr);
			}
			return me;
		}
	}

	/**
	 * A getInstance method for this singleton object that should be used to initialize the object.
	 * 
	 * @return a reference to the instace of the singleton object
	 * @va.exclude
	 */
	public static void createInstance() throws VistaLinkResourceException {

		if (logger.isEnabledFor(Priority.DEBUG)) {
			logger.debug("InstitutionMappingImpl :: In createInstance(institutionMap).");
		}

		synchronized (syncObjSingleton) {
			if (me == null) {
				if (logger.isEnabledFor(Priority.DEBUG)) {
					logger.debug("Instance is null, so creating.");
				}
				me = new InstitutionMappingImpl();
				// register singleton with Factory
				InstitutionMappingFactory.registerSingleton(me);
				logger.debug("Instance created.");

			} else {
				throw new VistaLinkResourceException("Institution Mapping has been already created.");
			}
		}
	}

	/**
	 * @return boolean indicating if instance already exists.
	 * @va.exclude
	 */
	public static boolean instanceExists() {
		synchronized (syncObjSingleton) {
			return (me != null);
		}
	}

	/**
	 * private constructor, to help enforce singleton characteristic.
	 */
	private InstitutionMappingImpl() {

		if (logger.isEnabledFor(Priority.DEBUG)) {
			logger.debug("InstitutionMappingImpl :: In private constructor.");
		}
		// create internal map of institutions
		institutionMap = new TreeMap();
	}

	/**
	 * Loads one or more station# mappings for a given JNDI name.
	 * 
	 * @param mcfJndiName
	 *            JNDI name to create mappings for
	 * @param mcfStationNumbers
	 *            One or more station#s to create mappings for
	 * @param mcfDistinguishedIdentifier
	 *            MCF distinguished identifier used later during a connector's finalize method -- allows a connector to
	 *            know whether it should remove the mapping or not, depending on a match on distinguished identifier.
	 * @va.exclude
	 */
	public void loadMappingsForJndiName(String mcfJndiName, String[] mcfStationNumbers, long mcfDistinguishedIdentifier)
			throws InstitutionMappingBadStationNumberException {

		// check station #s first
		for (int i = 0; i < mcfStationNumbers.length; i++) {
			checkStationNumber(mcfStationNumbers[i]);
		}

		synchronized (this.syncObjMapping) {

			// Remove any existing mappings for this jndi name that aren't part of the incoming station #s set
			Set keySet = this.institutionMap.keySet();
			for (Iterator iter = keySet.iterator(); iter.hasNext();) {
				String mappingStationNumber = (String) iter.next();
				String mappingJndiName = ((InstitutionMappingVO) this.institutionMap.get(mappingStationNumber))
						.getJndiName();
				if (mcfJndiName.equals(mappingJndiName)) {
					boolean stationMatch = false;
					for (int i = 0; i < mcfStationNumbers.length; i++) {
						if (mcfStationNumbers[i].equals(mappingStationNumber)) {
							stationMatch = true;
						}
					}
					if (!stationMatch) {
						if (logger.isDebugEnabled()) {
							logger.debug("removing mapping of station number '" + mappingStationNumber
									+ "' to JNDI name '" + mcfJndiName + "'.");
						}
						// remove the mapping
						iter.remove();
					}
				}
			}

			// Add new mappings:
			for (int i = 0; i < mcfStationNumbers.length; i++) {

				logger.debug("considering station#: " + mcfStationNumbers[i]);

				if (!institutionMap.containsKey(mcfStationNumbers[i])) {

					// If mapping not already present for station#, create new one
					if (logger.isDebugEnabled()) {
						logger.debug("adding mapping for JNDI name '" + mcfJndiName + "' to station number '"
								+ mcfStationNumbers[i] + "'.");
					}
					this.institutionMap.put(mcfStationNumbers[i], new InstitutionMappingVO(mcfJndiName,
							mcfDistinguishedIdentifier));

				} else {

					InstitutionMappingVO foundMapping = (InstitutionMappingVO) institutionMap.get(mcfStationNumbers[i]);
					String mappingJndiName = foundMapping.getJndiName();

					if (mappingJndiName.equals(mcfJndiName)) {

						// If mapping already present for this JNDI, reset MCF DI to new one
						if (logger.isDebugEnabled()) {
							logger
									.debug("Institution mapping contains station number: "
											+ mcfStationNumbers[i]
											+ "and the JNDI names are equal. Updating the last MCF distinguished identifier for mapping.");
						}
						foundMapping.setLastOneInMcfDistinguishedIdentifier(mcfDistinguishedIdentifier);

					} else {

						// If mapping already present for another JNDI, remove and create new one
						if (!mappingJndiName.equals(mcfJndiName)) {
							logger.warn("Existing station mapping found for station '" + mcfStationNumbers[i]
									+ "' mapped to JNDI name '" + mappingJndiName
									+ "'. Replacing with new mapping to JNDI name '" + mcfJndiName + "'.");
						}
						foundMapping.setJndiName(mcfJndiName);
						foundMapping.setLastOneInMcfDistinguishedIdentifier(mcfDistinguishedIdentifier);
					}
				}
			}
		}
	}

	/**
	 * apply business rules to station numbers
	 * 
	 * @param stationNumbers
	 *            array of one or more station numbers to check
	 * @throws InstitutionMappingBadStationNumberException
	 */
	private static void checkStationNumber(String stationNumber) throws InstitutionMappingBadStationNumberException {

		// - must be at least 3 digits
		// - if has alpha suffix, numeric portion must be "200"
		// - if numeric is 4 digits, 4th digit must be "9"

		if (logger.isDebugEnabled()) {
			logger.debug("looking up mapping for station # '" + stationNumber + "'.");
		}

		int nextDigitPosToTest = 0;
		while ((nextDigitPosToTest < stationNumber.length())
				&& (Character.isDigit(stationNumber.charAt(nextDigitPosToTest)))) {
			nextDigitPosToTest++;
		}

		//nextDigitPosToTest-1 has last successful position
		int lastDigitPos = nextDigitPosToTest - 1;

		if (lastDigitPos < 2) {
			// less than 3 digits -- not good
			// TODO confirm 3 digit limitation
			// TODO add unit test
			throw new InstitutionMappingBadStationNumberException("Station Number '" + stationNumber
					+ "' not valid; numeric portion must be at least 3 digits.");

		} else if ((lastDigitPos + 1) < stationNumber.length()) {
			// if non-numeric suffix, has to be 3 digits with first digit == 2
			if (lastDigitPos != 2) {
				throw new InstitutionMappingBadStationNumberException("Station Number '" + stationNumber
						+ "' not valid; only stations in the 200 range can have a non-numeric suffix.");
			} else if (stationNumber.charAt(0) != '2') {
				throw new InstitutionMappingBadStationNumberException("Station Number '" + stationNumber
						+ "' not valid; only stations in the 200 range can have a non-numeric suffix.");
			}

		} else if (lastDigitPos == 3) {
			// if 4 digits, check if 4th digit is '9'
			if (stationNumber.charAt(3) != '9') {
				throw new InstitutionMappingBadStationNumberException(
						"Station Number '"
								+ stationNumber
								+ "' not valid; if a 4th digit is present, it is treated as a suffix and can only be '9' for nursing homes.");
			}
		}
	}

	/**
	 * Call to remove mapping
	 * @param mcfJndiName
	 *            JNDI name of the mapping to remove
	 * @param mcfStationNumbers
	 *            Station # of the mapping to remove
	 * @param mcfDistinguishedIdentifier
	 *            Managed Connection Factory (mcf) distinguished identifier to identify owner of mapping to be removed.
	 * @va.exclude
	 */
	public void removeMappingsForJndiName(String mcfJndiName, String[] mcfStationNumbers,
			long mcfDistinguishedIdentifier) {

		synchronized (this.syncObjMapping) {

			for (int i = 0; i < mcfStationNumbers.length; i++) {
				logger.debug("considering removing station # " + mcfStationNumbers[i]);
				if (institutionMap.containsKey(mcfStationNumbers[i])) {
					InstitutionMappingVO imVO = (InstitutionMappingVO) institutionMap.get(mcfStationNumbers[i]);
					if (imVO.getJndiName().equals(mcfJndiName)) {
						if (imVO.getLastOneInMcfDistinguishedIdentifier() == mcfDistinguishedIdentifier) {
							institutionMap.remove(mcfStationNumbers[i]);
							if (logger.isDebugEnabled()) {
								logger.debug("Removed mapping for station #'" + mcfStationNumbers[i] + "', JNDI name '"
										+ mcfJndiName + "'.");
							}
						}
					}
				}
			}
		}
	}

	/**
	 * Returns the JNDI connector name (if found) for a division matching the station number passed in.
	 * 
	 * @param stationNumber
	 *            Station number of the division to retrieve a connector pool for.
	 * @return the JNDI connector name that has been mapped to the station number passed in.
	 * @throws InstitutionMapNotInitializedException
	 *             If the singleton instance of this class has not been created, this exception is thrown.
	 * @throws InsitutionMappingNotFoundException
	 *             if a match is not found, this exception is thrown.
	 * @va.exclude
	 */
	public String getJndiConnectorNameForInstitution(String stationNumber)
			throws InstitutionMapNotInitializedException, InstitutionMappingNotFoundException {
		InstitutionMappingVO imuVO = null;
		try {
			InstitutionMappingImpl me = getInstance();
			synchronized (this.syncObjMapping) {
				imuVO = (InstitutionMappingVO) me.institutionMap.get(stationNumber);
			}
			if (imuVO == null) {
				throw new InstitutionMappingNotFoundException("Could not match stationNumber '" + stationNumber
						+ "' to an institution mapping.");
			}
		} catch (IllegalStateException e) {
			throw new InstitutionMapNotInitializedException("The InstitutionMappingImpl object has not been created.");
		}
		return imuVO.getJndiName();
	}

	/**
	 * Used by MBean to display connector mapping list
	 * 
	 * @return a cloned copy of the TreeSet of current mappings
	 * @va.exclude
	 */
	public Map getMappingClone() {

		// make shallow copy of TreeMap, should prevent synch errors if real
		// TreeMap is being modified
		TreeMap clone = null;
		synchronized (this.syncObjMapping) {
			clone = (TreeMap) institutionMap.clone();
		}
		return clone;
	}

	/**
	 * Return a Set containing station#s in the current institution mapping.
	 * @return
	 */
	public Set getVistaLinkMappedStationNumberSet() {

		return Collections.unmodifiableSet(getMappingClone().keySet());

		//TODO commented out previous implementation 7/13/05 -- remove if new implementation above is satisfactory

		//		Map clone = getMappingClone();
		//		HashSet returnVal = new HashSet();
		//		Iterator iter = clone.keySet().iterator();
		//		String stationNum = null;
		//		while (iter.hasNext() ) {
		//			returnVal.add((String) iter.next());
		//		}
		//		return returnVal;
	}
}